# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-unary-operand-type, consider-using-generator

"""CNOT error density plot"""

import numpy as np
from scipy.stats import gaussian_kde
import matplotlib as mpl
import matplotlib.pyplot as plt
from qiskit.providers.models.backendproperties import BackendProperties
from kaleidoscope.colors import COLORS1, COLORS2, COLORS3, COLORS4, COLORS5, COLORS14
from kaleidoscope.errors import KaleidoscopeError
from kaleidoscope.qiskit.backends.pseudobackend import properties_to_pseudobackend


def cnot_error_density(backends,
                       figsize=None,
                       colors=None,
                       offset=None,
                       scale='log',
                       covariance_factor=0.1,
                       xlim=None,
                       text_xval=None,
                       xticks=None):
    """Plot CNOT error distribution for one or more IBMQ backends.

    Parameters:
        backends (list or IBMQBackend or BackendProperties): A single or ist of IBMQBackend
                                                             instances or properties.
        figsize (tuple): Optional figure size in inches.
        colors (list): A list of Matplotlib compatible colors to plot with.
        offset (float): Positive offset for spacing out the backends.
        scale (str): 'linear' or 'log' scaling of x-axis.
        covariance_factor (float): Sets the width of the gaussian for each point.
        xlim (list or tuple): Optional lower and upper limits of cnot error values.
        text_xval (float): Optional xaxis value at which to start the backend text.
        xticks (list): Optional list of xaxis ticks to plot.

    Returns:
        Figure: A matplotlib Figure instance.

    Raises:
        KaleidoscopeError: A backend with < 2 qubits was passed.
        KaleidoscopeError: Number of colors did not match number of backends.

    Example:
        .. jupyter-execute::

            from qiskit import *
            from kaleidoscope.qiskit.backends import cnot_error_density
            provider = IBMQ.load_account()

            backends = []
            backends.append(provider.backends.ibmq_vigo)
            backends.append(provider.backends.ibmq_ourense)
            backends.append(provider.backends.ibmq_valencia)
            backends.append(provider.backends.ibmq_santiago)

            cnot_error_density(backends)
    """

    if not isinstance(backends, list):
        backends = [backends]

    for idx, back in enumerate(backends):
        if isinstance(back, BackendProperties):
            backends[idx] = properties_to_pseudobackend(back)

    for back in backends:
        if back.configuration().n_qubits < 2:
            raise KaleidoscopeError('Number of backend qubits must be > 1')

    if scale not in ['linear', 'log']:
        raise KaleidoscopeError("scale must be 'linear' or 'log'.")

    # Attempt to autosize if figsize=None
    if figsize is None:
        if len(backends) > 1:
            fig = plt.figure(figsize=(12, len(backends)*1.5))
        else:
            fig = plt.figure(figsize=(12, 2))
    else:
        fig = plt.figure(figsize=figsize)

    text_color = 'k'
    if offset is None:
        offset = 100 if len(backends) > 3 else 200
    offset = -offset
    if colors is None:
        if len(backends) == 1:
            colors = COLORS1
        elif len(backends) == 2:
            colors = COLORS2
        elif len(backends) == 3:
            colors = COLORS3
        elif len(backends) == 4:
            colors = COLORS4
        elif len(backends) == 5:
            colors = COLORS5
        else:
            colors = [COLORS14[kk % 14] for kk in range(len(backends))]
    else:
        if len(colors) != len(backends):
            raise KaleidoscopeError('Number of colors does not match number of backends.')

    cx_errors = []
    for idx, back in enumerate(backends):

        back_props = back.properties().to_dict()

        cx_errs = []
        meas_errs = []
        for gate in back_props['gates']:
            if len(gate['qubits']) == 2:
                # Ignore cx gates with values of 1.0
                if gate['parameters'][0]['value'] != 1.0:
                    cx_errs.append(gate['parameters'][0]['value'])

        for qubit in back_props['qubits']:
            for item in qubit:
                if item['name'] == 'readout_error':
                    meas_errs.append(item['value'])
        cx_errors.append(100*np.asarray(cx_errs))

    max_cx_err = max([cerr.max() for cerr in cx_errors])
    min_cx_err = min([cerr.min() for cerr in cx_errors])

    if xlim is None:
        if scale == 'linear':
            xlim = [0, 1.5*max_cx_err]
        else:
            xlim = [10**np.floor(np.log10(min_cx_err)),
                    10**np.ceil(np.log10(max_cx_err))]

    if text_xval is None:
        if scale == 'linear':
            text_xval = 0.8*xlim[1]
        else:
            text_xval = 0.6*xlim[1]
    for idx, back in enumerate(backends):
        cx_density = gaussian_kde(cx_errors[idx])
        xs = np.linspace(xlim[0], xlim[1], 2500)
        cx_density.covariance_factor = lambda: covariance_factor
        cx_density._compute_covariance()

        if scale == 'linear':
            plt.plot(xs, 100*cx_density(xs)+offset*idx, zorder=idx, color=colors[idx])
        else:
            plt.semilogx(xs, 100*cx_density(xs)+offset*idx, zorder=idx, color=colors[idx])
        plt.fill_between(xs, offset*idx,
                         100*cx_density(xs)+offset*idx, zorder=idx, color=colors[idx])

        qv_val = back.configuration().quantum_volume
        if qv_val:
            qv = "(QV"+str(qv_val)+")"
        else:
            qv = ''

        bname = back.name().split('_')[-1].title()+" {}".format(qv)
        plt.text(text_xval, offset*idx+0.2*(-offset), bname, fontsize=20, color=colors[idx])

    fig.axes[0].get_yaxis().set_visible(False)
    # get rid of the frame
    for spine in plt.gca().spines.values():
        spine.set_visible(False)

    if xticks is None:
        if scale == 'linear':
            xticks = np.round(np.linspace(xlim[0], xlim[1], 4), 2)
    else:
        xticks = np.asarray(xticks)
    if xticks is not None:
        plt.xticks(np.floor(xticks), labels=np.floor(xticks), color=text_color)
    plt.xticks(fontsize=18)
    plt.xlim(xlim)
    plt.tick_params(axis='x', colors=text_color)
    plt.xlabel('Gate Error', fontsize=18, color=text_color)
    plt.title('CNOT Error Distributions', fontsize=18, color=text_color)
    fig.tight_layout()

    if mpl.get_backend() in ["module://ipykernel.pylab.backend_inline",
                             "module://matplotlib_inline.backend_inline",
                             "nbAgg"]:
        plt.close(fig)
    return fig



# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.

"""Routines for converting properties to pseudobackends for plotting."""


class Config():
    """Defines a just-good-enough configuration to use in plotting."""
    def __init__(self, name, num_qubits, cmap):
        """
        Parameters:
            name (str): Backend name.
            num_qubits (int): Number of qubits.
            cmap (list): Nested list for backend coupling map.
        """
        self.backend_name = name
        self.n_qubits = num_qubits
        self.coupling_map = cmap
        self.quantum_volume = None
        self.simulator = False


class PseudoBackend():
    """A pseudo-backend used for plotting passed properties."""
    def __init__(self, config, props):
        """
        Parameters:
            config (Config): A just enough configuration.
            props (BackendProperties): The backend properties.
        """
        self._config = config
        self._props = props

    def name(self):
        """Returns name of the backend.

        Returns:
            str: Backend name.
        """
        return self._config.backend_name

    def configuration(self):
        """Returns the backend configuration.

        Returns:
            Config: A just-good-enough configuration.
        """
        return self._config

    def properties(self):
        """Returns the backend properties.

        Returns:
            BackendProperties: The properties for a backend.
        """
        return self._props


def properties_to_pseudobackend(props):
    """Takes a BackendProperties object and makes a pseudobackend.

    Parameters:
        props (BackendProperties): The properties of a system.

    Returns:
        PseudoBackend: The corresponding pseudobackend.
    """
    name = props.backend_name
    num_qubits = len(props.qubits)
    cmap = []
    for gate in props.gates:
        if gate.gate == 'cx':
            cmap.append(gate.qubits)

    config = Config(name, num_qubits, cmap)

    return PseudoBackend(config, props)



# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=consider-using-generator

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=invalid-name

# Modified to also take backend parameters object, as well as the
# simulators objects from the Kaleidoscope providers.

"""Interactive error map for IBM Quantum Experience devices."""

import math
import numpy as np
import matplotlib as mpl
import plotly.graph_objects as go
from plotly.subplots import make_subplots
from qiskit.providers.models.backendproperties import BackendProperties
from qiskit.providers.ibmq.ibmqbackend import IBMQBackend
from qiskit.providers.fake_provider import FakeBackend
from kaleidoscope.errors import KaleidoscopeError
from kaleidoscope.colors.utils import find_text_color
from kaleidoscope.interactive.plotly_wrapper import PlotlyWidget, PlotlyFigure
from kaleidoscope.qiskit.backends.device_layouts import LAYOUTS
from kaleidoscope.qiskit.backends.pseudobackend import properties_to_pseudobackend
from kaleidoscope.colors import BMW
from kaleidoscope.colors.cmap import cmap_to_plotly


def system_error_map(backend,
                     figsize=(None, None),
                     colormap=None,
                     background_color='white',
                     show_title=True,
                     remove_badcal_edges=True,
                     as_widget=False):
    """Plot the error map of a device.

    Args:
        backend (IBMQBackend or FakeBackend or DeviceSimulator or Properties): Plot the error map
                                                                               for a backend.
        figsize (tuple, optional): Figure size in pixels.
        colormap (Colormap): A matplotlib colormap.
        background_color (str, optional): Background color, either 'white' or 'black'.
        show_title (bool, optional): Whether to show figure title.
        remove_badcal_edges (bool, optional): Whether to remove bad CX gate calibration data.
        as_widget (bool, optional): ``True`` if the figure is to be returned as a ``PlotlyWidget``.
            Otherwise the figure is to be returned as a ``PlotlyFigure``.

    Returns:
        PlotlyFigure or PlotlyWidget: The error map figure.

    Raises:
        KaleidoscopeError: Invalid input type.

    Example:
        .. jupyter-execute::

            from qiskit import *
            from kaleidoscope.qiskit.backends import system_error_map

            pro = IBMQ.load_account()
            backend = pro.backends.ibmq_vigo
            system_error_map(backend)

    """
    if not isinstance(backend, (IBMQBackend, FakeBackend, BackendProperties)):
        raise KaleidoscopeError('Input is not a valid backend or properties object.')

    if isinstance(backend, BackendProperties):
        backend = properties_to_pseudobackend(backend)

    CMAP = BMW
    PLOTLY_CMAP = cmap_to_plotly(CMAP)

    if colormap is not None:
        CMAP = colormap
        PLOTLY_CMAP = cmap_to_plotly(CMAP)

    meas_text_color = '#000000'
    if background_color == 'white':
        color_map = CMAP
        text_color = '#000000'
        plotly_cmap = PLOTLY_CMAP
    elif background_color == 'black':
        color_map = CMAP
        text_color = '#FFFFFF'
        plotly_cmap = PLOTLY_CMAP
    else:
        raise KaleidoscopeError(
            '"{}" is not a valid background_color selection.'.format(background_color))

    if backend.configuration().simulator:
        raise KaleidoscopeError('Requires a device backend, not a simulator.')

    config = backend.configuration()
    n_qubits = config.n_qubits
    cmap = config.coupling_map

    if str(n_qubits) in LAYOUTS['layouts'].keys():
        kind = 'generic'
        if backend.name() in LAYOUTS['special_names']:
            if LAYOUTS['special_names'][backend.name()] in LAYOUTS['layouts'][str(n_qubits)]:
                kind = LAYOUTS['special_names'][backend.name()]
        grid_data = LAYOUTS['layouts'][str(n_qubits)][kind]
    else:
        fig = go.Figure()
        fig.update_layout(showlegend=False,
                          plot_bgcolor=background_color,
                          paper_bgcolor=background_color,
                          width=figsize[0], height=figsize[1],
                          margin=dict(t=60, l=0, r=0, b=0)
                          )
        out = PlotlyWidget(fig)
        return out

    props = backend.properties()

    freqs = [0] * n_qubits
    t1s = [0] * n_qubits
    t2s = [0] * n_qubits
    alphas = [0] * n_qubits
    for idx, qubit_props in enumerate(props.qubits):
        for item in qubit_props:
            if item.name == 'frequency':
                freqs[idx] = item.value
            elif item.name == 'T1':
                t1s[idx] = item.value
            elif item.name == 'T2':
                t2s[idx] = item.value
            elif item.name == 'anharmonicity':
                alphas[idx] = item.value

    # U2 error rates
    single_gate_errors = [0]*n_qubits
    single_gate_times = [0]*n_qubits
    for gate in props.gates:
        if gate.gate in ['u2', 'sx']:
            _qubit = gate.qubits[0]
            for gpar in gate.parameters:
                if gpar.name == 'gate_error':
                    single_gate_errors[_qubit] = gpar.value
                elif gpar.name == 'gate_length':
                    single_gate_times[_qubit] = gpar.value

    # Convert to log10
    single_gate_errors = np.log10(np.asarray(single_gate_errors))

    avg_1q_err = np.mean(single_gate_errors)
    max_1q_err = _round_log10_exp(np.max(single_gate_errors), rnd='up', decimals=1)
    min_1q_err = _round_log10_exp(np.min(single_gate_errors), rnd='down', decimals=1)

    single_norm = mpl.colors.Normalize(vmin=min_1q_err, vmax=max_1q_err)

    q_colors = [mpl.colors.rgb2hex(color_map(single_norm(err))) for err in single_gate_errors]

    if n_qubits > 1:
        line_colors = []
        if cmap:
            cx_errors = []
            cx_times = []
            for line in cmap:
                for gate in props.gates:
                    if gate.qubits == line:
                        for gpar in gate.parameters:
                            if gpar.name == 'gate_error':
                                cx_errors.append(gpar.value)
                            elif gpar.name == 'gate_length':
                                cx_times.append(gpar.value)

            # Convert to array
            cx_errors = np.log10(np.asarray(cx_errors))

            # remove bad cx edges
            if remove_badcal_edges:
                cx_idx = np.where(cx_errors != 0.0)[0]
            else:
                cx_idx = np.arange(len(cx_errors))

            avg_cx_err = np.mean(cx_errors[cx_idx])
            min_cx_err = _round_log10_exp(np.min(cx_errors[cx_idx]), rnd='down', decimals=1)
            max_cx_err = _round_log10_exp(np.max(cx_errors[cx_idx]), rnd='up', decimals=1)

            cx_norm = mpl.colors.Normalize(vmin=min_cx_err, vmax=max_cx_err)

            for err in cx_errors:
                if err != 0.0 or not remove_badcal_edges:
                    line_colors.append(mpl.colors.rgb2hex(color_map(cx_norm(err))))
                else:
                    line_colors.append("#ff0000")

    # Measurement errors
    read_err = [0] * n_qubits
    p01_err = [0] * n_qubits
    p10_err = [0] * n_qubits
    for qubit in range(n_qubits):
        for item in props.qubits[qubit]:
            if item.name == 'readout_error':
                read_err[qubit] = item.value
            elif item.name == 'prob_meas0_prep1':
                p01_err[qubit] = item.value
            elif item.name == 'prob_meas1_prep0':
                p10_err[qubit] = item.value

    read_err = np.asarray(read_err)
    avg_read_err = np.mean(read_err)
    max_read_err = np.max(read_err)
    p01_err = np.asarray(p01_err)
    p10_err = np.asarray(p10_err)

    if n_qubits < 10:
        num_left = n_qubits
        num_right = 0
    else:
        num_left = math.ceil(n_qubits / 2)
        num_right = n_qubits - num_left

    x_max = max([d[1] for d in grid_data])
    y_max = max([d[0] for d in grid_data])
    max_dim = max(x_max, y_max)

    qubit_size = 32
    font_size = 14
    offset = 0
    if cmap:
        if y_max / max_dim < 0.33:
            qubit_size = 24
            font_size = 10
            offset = 1

    if n_qubits > 5:
        right_meas_title = "Readout error"
    else:
        right_meas_title = None

    if cmap:
        cx_title = "CNOT error rate [Avg. {}]".format('{:.2}\u22C510<sup>{}</sup>'.format(
            *_pow10_coeffs(avg_cx_err)))
    else:
        cx_title = None
    fig = make_subplots(rows=2, cols=11, row_heights=[0.95, 0.05],
                        vertical_spacing=0.15,
                        specs=[[{"colspan": 2}, None, {"colspan": 6},
                                None, None, None,
                                None, None, {"colspan": 2},
                                None, None],
                               [{"colspan": 4}, None, None,
                                None, None, None,
                                {"colspan": 4}, None, None,
                                None, None]],
                        subplot_titles=("Readout error", None, right_meas_title,
                                        "SX error rate [Avg. {}]".format(
                                            '{:.2}\u22C510<sup>{}</sup>'.format(
                                                *_pow10_coeffs(avg_1q_err))),
                                        cx_title)
                        )

    # Add lines for couplings
    if cmap and n_qubits > 1:
        for ind, edge in enumerate(cmap):
            is_symmetric = False
            if edge[::-1] in cmap:
                is_symmetric = True
            y_start = grid_data[edge[0]][0] + offset
            x_start = grid_data[edge[0]][1]
            y_end = grid_data[edge[1]][0] + offset
            x_end = grid_data[edge[1]][1]

            if is_symmetric:
                if y_start == y_end:
                    x_end = (x_end - x_start) / 2 + x_start
                    x_mid = x_end
                    y_mid = y_start

                elif x_start == x_end:
                    y_end = (y_end - y_start) / 2 + y_start
                    x_mid = x_start
                    y_mid = y_end

                else:
                    x_end = (x_end - x_start) / 2 + x_start
                    y_end = (y_end - y_start) / 2 + y_start
                    x_mid = x_end
                    y_mid = y_end
            else:
                if y_start == y_end:
                    x_mid = (x_end - x_start) / 2 + x_start
                    y_mid = y_end

                elif x_start == x_end:
                    x_mid = x_end
                    y_mid = (y_end - y_start) / 2 + y_start

                else:
                    x_mid = (x_end - x_start) / 2 + x_start
                    y_mid = (y_end - y_start) / 2 + y_start

            cx_str = 'cnot<sub>err</sub> = {err}'
            cx_str += '<br>&#120591;<sub>cx</sub>     = {tau} ns'
            fig.append_trace(
                go.Scatter(x=[x_start, x_mid, x_end],
                           y=[-y_start, -y_mid, -y_end],
                           mode="lines",
                           line=dict(width=6,
                                     color=line_colors[ind]),
                           hoverinfo='text',
                           hovertext=cx_str.format(
                               err='{:.3}\u22C510<sup>{}</sup>'.format(
                                   *_pow10_coeffs(cx_errors[ind])),
                               tau=np.round(cx_times[ind], 2))
                           ),
                row=1, col=3)

    # Add the qubits themselves
    qubit_text = []
    qubit_str = "<b>Qubit {idx}</b>"
    qubit_str += "<br>freq = {freq} GHz"
    qubit_str += "<br>T<sub>1</sub>   = {t1} \u03BCs"
    qubit_str += "<br>T<sub>2</sub>   = {t2} \u03BCs"
    qubit_str += "<br>&#945;    = {anh} GHz"
    qubit_str += "<br>sx<sub>err</sub> = {err}"
    qubit_str += "<br>&#120591;<sub>sx</sub>   = {tau} ns"
    for kk in range(n_qubits):
        qubit_text.append(qubit_str.format(idx=kk,
                                           freq=np.round(freqs[kk], 5),
                                           t1=np.round(t1s[kk], 2),
                                           t2=np.round(t2s[kk], 2),
                                           anh=np.round(alphas[kk], 3) if alphas[kk] else 'NA',
                                           err='{:.3}\u22C510<sup>{}</sup>'.format(
                                               *_pow10_coeffs(single_gate_errors[kk])),
                                           tau=np.round(single_gate_times[kk], 2)))

    if n_qubits > 20:
        qubit_size = 23
        font_size = 11

    if n_qubits > 50:
        qubit_size = 20
        font_size = 9

    qtext_color = []
    for ii in range(n_qubits):
        qtext_color.append(find_text_color(q_colors[ii]))

    fig.append_trace(go.Scatter(
        x=[d[1] for d in grid_data],
        y=[-d[0]-offset for d in grid_data],
        mode="markers+text",
        marker=go.scatter.Marker(size=qubit_size,
                                 color=q_colors,
                                 opacity=1),
        text=[str(ii) for ii in range(n_qubits)],
        textposition="middle center",
        textfont=dict(size=font_size, color=qtext_color),
        hoverinfo="text",
        hovertext=qubit_text), row=1, col=3)

    fig.update_xaxes(row=1, col=3, visible=False)
    _range = None
    if offset:
        _range = [-3.5, 0.5]
    fig.update_yaxes(row=1,
                     col=3,
                     visible=False,
                     range=_range)

    # H error rate colorbar
    if n_qubits > 1:
        fig.append_trace(go.Heatmap(z=[np.linspace(min_1q_err,
                                                   max_1q_err, 100),
                                       np.linspace(min_1q_err,
                                                   max_1q_err, 100)],
                                    colorscale=plotly_cmap,
                                    showscale=False,
                                    hoverinfo='none'), row=2, col=1)

        fig.update_yaxes(row=2,
                         col=1,
                         visible=False)

        mid_1q_err = _round_log10_exp((max_1q_err-min_1q_err)/2+min_1q_err,
                                      rnd='up', decimals=1)

        fig.update_xaxes(row=2,
                         col=1,
                         tickfont=dict(size=13),
                         tickvals=[0, 49, 99],
                         ticktext=['{:.2}\u22C510<sup>{}</sup>'.format(
                             *_pow10_coeffs(min_1q_err)),
                                   '{:.2}\u22C510<sup>{}</sup>'.format(
                                       *_pow10_coeffs(mid_1q_err)),
                                   '{:.2}\u22C510<sup>{}</sup>'.format(
                                       *_pow10_coeffs(max_1q_err)),
                                   ])

    # CX error rate colorbar
    if cmap and n_qubits > 1:
        fig.append_trace(go.Heatmap(z=[np.linspace(min_cx_err,
                                                   max_cx_err, 100),
                                       np.linspace(min_cx_err,
                                                   max_cx_err, 100)],
                                    colorscale=plotly_cmap,
                                    showscale=False,
                                    hoverinfo='none'), row=2, col=7)

        fig.update_yaxes(row=2, col=7, visible=False)

        mid_cx_err = (max_cx_err-min_cx_err)/2 + min_cx_err
        fig.update_xaxes(row=2, col=7,
                         tickfont=dict(size=13),
                         tickvals=[0, 49, 99],
                         ticktext=['{:.2}\u22C510<sup>{}</sup>'.format(
                             *_pow10_coeffs(min_cx_err)),
                                   '{:.2}\u22C510<sup>{}</sup>'.format(
                                       *_pow10_coeffs(mid_cx_err)),
                                   '{:.2}\u22C510<sup>{}</sup>'.format(
                                       *_pow10_coeffs(max_cx_err))])

    hover_text = "<b>Qubit {idx}</b>"
    hover_text += "<br>M<sub>err</sub> = {err}"
    hover_text += "<br>P<sub>0|1</sub> = {p01}"
    hover_text += "<br>P<sub>1|0</sub> = {p10}"
    # Add the left side meas errors
    for kk in range(num_left-1, -1, -1):
        fig.append_trace(go.Bar(x=[read_err[kk]], y=[kk],
                                orientation='h',
                                marker=dict(color='#c7c7c5'),
                                hoverinfo="text",
                                hoverlabel=dict(font=dict(color=meas_text_color)),
                                hovertext=[hover_text.format(idx=kk,
                                                             err=np.round(read_err[kk], 4),
                                                             p01=np.round(p01_err[kk], 4),
                                                             p10=np.round(p10_err[kk], 4)
                                                             )]
                                ),
                         row=1, col=1)

    fig.append_trace(go.Scatter(x=[avg_read_err, avg_read_err],
                                y=[-0.25, num_left-1+0.25],
                                mode='lines',
                                hoverinfo='none',
                                line=dict(color=text_color,
                                          width=2,
                                          dash='dot')), row=1, col=1)

    fig.update_yaxes(row=1, col=1,
                     tickvals=list(range(num_left)),
                     autorange="reversed")

    fig.update_xaxes(row=1, col=1,
                     range=[0, 1.1*max_read_err],
                     tickvals=[0, np.round(avg_read_err, 2),
                               np.round(max_read_err, 2)],
                     showline=True, linewidth=1, linecolor=text_color,
                     tickcolor=text_color,
                     ticks="outside",
                     showgrid=False,
                     zeroline=False)

    # Add the right side meas errors, if any
    if num_right:
        for kk in range(n_qubits-1, num_left-1, -1):
            fig.append_trace(go.Bar(x=[-read_err[kk]],
                                    y=[kk],
                                    orientation='h',
                                    marker=dict(color='#c7c7c5'),
                                    hoverinfo="text",
                                    hoverlabel=dict(font=dict(color=meas_text_color)),
                                    hovertext=[hover_text.format(idx=kk,
                                                                 err=np.round(read_err[kk], 4),
                                                                 p01=np.round(p01_err[kk], 4),
                                                                 p10=np.round(p10_err[kk], 4)
                                                                 )
                                               ]
                                    ),
                             row=1, col=9)

        fig.append_trace(go.Scatter(x=[-avg_read_err, -avg_read_err],
                                    y=[num_left-0.25, n_qubits-1+0.25],
                                    mode='lines',
                                    hoverinfo='none',
                                    line=dict(color=text_color,
                                              width=2,
                                              dash='dot')
                                    ), row=1, col=9)

        fig.update_yaxes(row=1,
                         col=9,
                         tickvals=list(range(n_qubits-1, num_left-1, -1)),
                         side='right',
                         autorange="reversed",
                         )

        fig.update_xaxes(row=1,
                         col=9,
                         range=[-1.1*max_read_err, 0],
                         tickvals=[0, -np.round(avg_read_err, 2), -np.round(max_read_err, 2)],
                         ticktext=[0, np.round(avg_read_err, 2), np.round(max_read_err, 2)],
                         showline=True, linewidth=1, linecolor=text_color,
                         tickcolor=text_color,
                         ticks="outside",
                         showgrid=False,
                         zeroline=False)

    # Makes the subplot titles smaller than the 16pt default
    for ann in fig['layout']['annotations']:
        ann['font'] = dict(size=13)

    title_text = "{} error map".format(backend.name()) if show_title else ''
    fig.update_layout(showlegend=False,
                      plot_bgcolor=background_color,
                      paper_bgcolor=background_color,
                      width=figsize[0], height=figsize[1],
                      title=dict(text=title_text, x=0.452),
                      title_font_size=20,
                      font=dict(color=text_color),
                      margin=dict(t=60, l=0, r=0, b=0),
                      hoverlabel=dict(font_size=14,
                                      font_family="courier,monospace",
                                      align='left'
                                      )
                      )
    if as_widget:
        return PlotlyWidget(fig)
    return PlotlyFigure(fig)


def _round_up(n, decimals=0):
    multiplier = 10**decimals
    return np.ceil(n*multiplier) / multiplier


def _round_down(n, decimals=0):
    multiplier = 10**decimals
    return np.floor(n*multiplier) / multiplier


def _pow10_coeffs(x):
    """Returns the coefficient A of a number A*10**y
    as well as the exponent y.

    Parameters:
        x (float): Input number in log10.

    Returns:
        float: Normed value
        int: The exponent.
    """
    z = abs(x) + (1 if x < 0 else 0)
    y = (-1*np.sign(x)*np.floor(z))
    return (10**x)*10**y, int(-y)


def _round_log10_exp(x, rnd='up', decimals=1):
    """Rounds a log10 number to the nearest value
    such that it can be cleanly written as A*10**y

    Parameters:
        x (float): Input number in log10.
        rnd (str): Round 'up' or 'down'.
        decimals (int): Number of decimals to round to.

    Returns:
        float: The adjusted value log10.
    """
    normed_value, y = _pow10_coeffs(x)
    if rnd == 'up':
        normed_value = _round_up(normed_value, decimals=decimals)
    elif rnd == 'down':
        normed_value = _round_down(normed_value, decimals=decimals)
    return np.log10(normed_value)+y



# -*- coding: utf-8 -*-

# This code is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Interactive Bloch discs"""

import math
import numpy as np
import plotly.graph_objects as go
from plotly.subplots import make_subplots
import matplotlib.pyplot as plt
import matplotlib.colors

from kaleidoscope.utils import pi_check
from kaleidoscope.interactive.plotly_wrapper import PlotlyFigure, PlotlyWidget
from kaleidoscope.interactive.bloch.utils import bloch_components
from kaleidoscope.colors.utils import hex_to_rgb
from kaleidoscope.colors import BJY
from kaleidoscope.colors.cmap import cmap_to_plotly

NORM = plt.Normalize(-1, 1)


def bloch_sunburst(vec, colormap):
    """Create a Bloch disc using a Plotly sunburst.

    Parameters:
        vec (ndarray): A vector of Bloch components.
        colormap (Colormap): A matplotlib colormap.

    Returns:
        go.Figure: A Plotly figure instance,

    Raises:
        ValueError: Input vector is not normalized.
    """
    eps = 1e-6
    vec = np.asarray(vec)
    vec_norm = np.linalg.norm(vec)
    if vec_norm > 1.0 + eps:
        raise ValueError('Input vector has length {} greater than 1.0'.format(vec_norm))

    for idx, val in enumerate(vec):
        if abs(val) < 1e-15:
            vec[idx] = 0

    th = math.atan2(vec[1], vec[0])

    if th < 0:
        th = 2*np.pi+th

    z_hex = matplotlib.colors.rgb2hex(colormap(NORM(vec[2])))

    z_color = "rgba({},{},{},{})".format(*hex_to_rgb(z_hex), 0.95*vec_norm+0.05)
    ring_color = "rgba({},{},{},{})".format(*hex_to_rgb('#000000'), 0.95*vec_norm+0.05)

    wedge_str = "\u2329X\u232A= {x}<br>"
    wedge_str += "\u2329Y\u232A= {y}<br>"
    wedge_str += "\u2329Z\u232A= {z}<br>"
    wedge_str += " \u03B8  = {th}<br>"
    wedge_str += "|\u03C8| = {pur}"

    th_str = pi_check(th, ndigits=3)
    th_str = th_str.replace('pi', '\u03C0')

    hover_text = [wedge_str.format(x=round(vec[0], 3),
                                   y=round(vec[1], 3),
                                   z=round(vec[2], 3),
                                   th=th_str,
                                   pur=np.round(vec_norm, 3))] + [None]

    bloch = go.Sunburst(labels=[" ", "  "],
                        parents=["", " "],
                        values=[2*np.pi-th, th],
                        hoverinfo="text",
                        hovertext=hover_text,
                        marker=dict(colors=[z_color, ring_color]))
    return bloch


def bloch_disc(rho, figsize=None, title=None, colormap=None, as_widget=False):
    """Plot a Bloch disc for a single qubit.

    Parameters:
        rho (list or ndarray or Statevector or DensityMatrix): Input statevector, density matrix,
                                                               or Bloch components.
        figsize (tuple): Figure size in pixels, default=(200,275).
        title (str): Plot title.
        colormap (Colormap): A matplotlib colormap.
        as_widget (bool): Return plot as a widget.

    Returns:
        PlotlyFigure: A Plotly figure instance
        PlotlyWidget : A Plotly widget if `as_widget=True`.

    Example:
        .. jupyter-execute::

            import numpy as np
            from qiskit import *
            from qiskit.quantum_info import Statevector

            from kaleidoscope.interactive import bloch_disc
            qc = QuantumCircuit(1)
            qc.ry(np.pi*np.random.random(), 0)
            qc.rz(np.pi*np.random.random(), 0)

            state = Statevector.from_instruction(qc)
            bloch_disc(state)

    """
    # A hack so I do not have to import the actual instances from Qiskit.
    if rho.__class__.__name__ in ['Statevector', 'DensityMatrix'] \
            and 'qiskit' in rho.__class__.__module__:
        rho = rho.data
    if len(rho) != 3:
        rho = np.asarray(rho, dtype=complex)
        comp = bloch_components(rho)
    else:
        comp = [rho]

    if title:
        title = [title] + ["\u2329Z\u232A"]
    else:
        title = [""] + ["\u2329Z\u232A"]

    if colormap is None:
        colormap = BJY

    if figsize is None:
        figsize = (200, 275)

    fig = make_subplots(rows=1, cols=2,
                        specs=[[{'type': 'domain'}]+[{'type': 'xy'}]],
                        subplot_titles=title,
                        column_widths=[0.93]+[0.07])

    fig.add_trace(bloch_sunburst(comp[0], colormap), row=1, col=1)

    zval = comp[0][2]
    zrange = [k*np.ones(1) for k in np.linspace(-1, 1, 100)]

    idx = (np.abs(np.linspace(-1, 1, 100) - zval)).argmin()

    tickvals = np.array([0, 49, 99, idx])
    idx_sort = np.argsort(tickvals)
    tickvals = tickvals[idx_sort]

    ticktext = [-1, 0, 1, "\u25C0"+str(np.round(zval, 3))]
    if zval <= -0.95:
        ticktext[0] = ''
    elif abs(zval) <= 0.05:
        ticktext[1] = ''
    elif zval >= 0.95:
        ticktext[2] = ''

    ticktext = [ticktext[kk] for kk in idx_sort]

    PLOTLY_CMAP = cmap_to_plotly(colormap)
    fig.append_trace(go.Heatmap(z=zrange,
                                colorscale=PLOTLY_CMAP,
                                showscale=False,
                                hoverinfo='none',
                                ),
                     row=1, col=2
                     )

    fig.update_yaxes(row=1, col=2, tickvals=tickvals,
                     ticktext=ticktext)

    fig.update_yaxes(row=1, col=2, side="right")
    fig.update_xaxes(row=1, col=2, visible=False)

    fig.update_layout(margin=dict(t=30, l=10, r=0, b=0),
                      height=figsize[0],
                      width=figsize[1],
                      hoverlabel=dict(font_size=16,
                                      font_family="courier,monospace",
                                      align='left'
                                      )
                      )
    for ann in fig['layout']['annotations']:
        ann['font'] = dict(size=14)

    if as_widget:
        return PlotlyWidget(fig)

    return PlotlyFigure(fig)


def bloch_multi_disc(rho, figsize=None, titles=True, colormap=None, as_widget=False):
    """Plot Bloch discs for a multi-qubit state.

    Parameters:
        rho (list or ndarray or Statevector or DensityMatrix): Input statevector, density matrix.
        figsize (tuple): Figure size in pixels, default=(125*num_qubits, 150).
        titles (bool): Display titles.
        colormap (Colormap): A matplotlib colormap.
        as_widget (bool): Return plot as a widget.

    Returns:
        PlotlyFigure: A Plotly figure instance
        PlotlyWidget : A Plotly widget if `as_widget=True`.

    Example:
        .. jupyter-execute::

            import numpy as np
            from qiskit import *
            from qiskit.quantum_info import Statevector
            from kaleidoscope.interactive import bloch_multi_disc

            N = 4
            qc = QuantumCircuit(N)
            qc.h(range(N))
            for kk in range(N):
                qc.ry(2*np.pi*np.random.random(), kk)
            for kk in range(N-1):
                qc.cx(kk,kk+1)
            for kk in range(N):
                qc.rz(2*np.pi*np.random.random(), kk)

            state = Statevector.from_instruction(qc)
            bloch_multi_disc(state)
    """
    # A hack so I do not have to import the actual instances from Qiskit.
    if rho.__class__.__name__ in ['Statevector', 'DensityMatrix'] \
            and 'qiskit' in rho.__class__.__module__:
        rho = rho.data

    rho = np.asarray(rho, dtype=complex)

    comp = bloch_components(rho)
    num = int(np.log2(rho.shape[0]))

    nrows = 1
    ncols = num

    if figsize is None:
        figsize = (ncols*125, 150)

    if titles:
        titles = ["Qubit {}".format(k) for k in range(num)] + ["\u2329Z\u232A"]
    else:
        titles = ["" for k in range(num)] + ["\u2329Z\u232A"]

    if colormap is None:
        colormap = BJY

    fig = make_subplots(rows=nrows, cols=ncols+1,
                        specs=[[{'type': 'domain'}]*ncols+[{'type': 'xy'}]],
                        subplot_titles=titles,
                        column_widths=[0.95/num]*num+[0.05])

    for jj in range(num):
        fig.add_trace(bloch_sunburst(comp[jj], colormap), row=1, col=jj+1)

    zrange = [k*np.ones(1) for k in np.linspace(-1, 1, 100)]
    PLOTLY_CMAP = cmap_to_plotly(colormap)
    fig.append_trace(go.Heatmap(z=zrange,
                                colorscale=PLOTLY_CMAP,
                                showscale=False,
                                hoverinfo='none',
                                ),
                     row=1, col=num+1)

    fig.update_yaxes(row=1, col=num+1, tickvals=[0, 49, 99],
                     ticktext=[-1, 0, 1])
    fig.update_yaxes(row=1, col=num+1, side="right")
    fig.update_xaxes(row=1, col=num+1, visible=False)

    fig.update_layout(margin=dict(t=50, l=0, r=15, b=30),
                      width=figsize[0],
                      height=figsize[1],
                      hoverlabel=dict(font_size=14,
                                      font_family="monospace",
                                      align='left'
                                      )
                      )
    # Makes the subplot titles smaller than the 16pt default
    for ann in fig['layout']['annotations']:
        ann['font'] = dict(size=16)

    if as_widget:
        return PlotlyWidget(fig)
    return PlotlyFigure(fig)



# -*- coding: utf-8 -*-

# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=invalid-name
# Configuration file for the Sphinx documentation builder.
#
# This file does only contain a selection of the most common options. For a
# full list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))

import kaleidoscope as kal

"""
Sphinx documentation builder
"""

# The short X.Y version
version = kal.__version__
# The full version, including alpha/beta/rc tags
release = kal.__version__


rst_prolog = """
.. |version| replace:: {0}
""".format(kal.version.short_version)

# -- Project information -----------------------------------------------------
project = 'Kaleidoscope {}'.format(version)
copyright = '2021, Kaleidoscope Team'  # pylint: disable=redefined-builtin
author = 'Kaleidoscope Development Team'
# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.napoleon',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.mathjax',
    'sphinx.ext.viewcode',
    'sphinx.ext.extlinks',
    'jupyter_sphinx'
]
html_static_path = ['_static']
templates_path = ['_templates']
html_css_files = ['gallery.css']

exclude_patterns = ['_build', '**.ipynb_checkpoints']

# -----------------------------------------------------------------------------
# Autosummary
# -----------------------------------------------------------------------------
autosummary_generate = True

# -----------------------------------------------------------------------------
# Autodoc
# -----------------------------------------------------------------------------

autodoc_default_options = {
    'inherited-members': None,
}

autoclass_content = 'both'


# If true, figures, tables and code-blocks are automatically numbered if they
# have a caption.
numfig = True

# A dictionary mapping 'figure', 'table', 'code-block' and 'section' to
# strings that are used for format of figure numbers. As a special character,
# %s will be replaced to figure number.
numfig_format = {
    'table': 'Table %s'
}
# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# A boolean that decides whether module names are prepended to all object names
# (for object types where a “module” of some kind is defined), e.g. for
# py:function directives.
add_module_names = False

# A list of prefixes that are ignored for sorting the Python module index
# (e.g., if this is set to ['foo.'], then foo.bar is shown under B, not F).
# This can be handy if you document a project that consists of a single
# package. Works only for the HTML builder currently.
modindex_common_prefix = ['kaleidoscope.']

# -- Configuration for extlinks extension ------------------------------------
# Refer to https://www.sphinx-doc.org/en/master/usage/extensions/extlinks.html


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "pydata_sphinx_theme"

html_logo = 'images/kal_logo.png'

html_theme_options = {
    "github_url": "https://github.com/nonhermitian/kaleidoscope",
    "use_edit_page_button": True,
}

html_context = {
    "github_user": "nonhermitian",
    "github_repo": "kaleidoscope",
    "github_version": "master",
    "doc_path": "docs",
}
#html_sidebars = {'**': ['globaltoc.html']}
html_last_updated_fmt = '%Y/%m/%d'



# -*- coding: utf-8 -*-

# This code is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Colormap routines"""
import numpy as np


def cmap_to_plotly(cmap):
    """Convert a color map to a Plotly RGBA color scale.

    Parameters:
        cmap (matplotlib.colors.Colormap): Color map to be converted.

    Returns:
        list: RGBA Colorscale.
    """
    pl_entries = 255
    hgt = 1.0/(pl_entries-1)
    pl_colorscale = []

    for k in range(pl_entries):
        clr = list(map(np.uint8, np.array(cmap(k*hgt)[:3])*255))
        pl_colorscale.append([k*hgt, 'rgba'+str((clr[0], clr[1], clr[2], 1.0))])

    return pl_colorscale



# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
# pylint: disable=wrong-import-position, unused-argument, unused-import

"""Kaleidoscope"""

import os
from kaleidoscope.errors import KaleidoscopeError

# This is needed because version info is only generated
# at setup.  This should only fall back when not using
# setup.py lint or style to check.
try:
    from .version import version as __version__
except ImportError:
    __version__ = '0.0.0'

from kaleidoscope.interactive import *

try:
    from qiskit import QuantumCircuit
    from qiskit.providers.aer import Aer
    from qiskit.providers.ibmq import IBMQ
except ImportError:
    HAS_QISKIT = False
else:
    HAS_QISKIT = True



# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=wrong-import-position

"""Qiskit specific functionality"""

from kaleidoscope import HAS_QISKIT
from kaleidoscope.errors import KaleidoscopeError

if not HAS_QISKIT:
    raise KaleidoscopeError('Must install qiskit-terra, qiskit-aer, and qiskit-ibmq-provider.')

from .backends.mpl import *
from .backends.interactive import system_error_map, system_gate_map



# -*- coding: utf-8 -*-

# This code is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Interactive plotting utility functions"""

import numpy as np
import scipy.sparse as sp
from numba import vectorize, uint32, int32, complex128, jit, prange


@vectorize([uint32(uint32)], target='parallel', nopython=True, cache=True)
def count_set_bits(val):
    """Computes the number of set bits in a uint32 value.

    Parameters:
        val (uint32): Input unsigned int.

    Returns:
        uint32: Output unsigned int.
    """
    val = (val & 0x55555555) + ((val >> 1) & 0x55555555)
    val = (val & 0x33333333) + ((val >> 2) & 0x33333333)
    val = (val & 0x0f0f0f0f) + ((val >> 4) & 0x0f0f0f0f)
    val = (val & 0x00ff00ff) + ((val >> 8) & 0x00ff00ff)
    val = (val & 0x0000ffff) + ((val >> 16) & 0x0000ffff)
    return val


@jit(complex128(complex128[:], int32[:], int32[:], complex128[:]),
     nopython=True, cache=True)
def expect_psi_csr(data, ind, ptr, vec):
    """Computes the expectation value of a complex matrix in CSR sparse format.

    Note that this routine returns a complex type regardless of whether
    the matrix is Hermitian or not.  If so, take the real part.

    Parameters:
        data (ndarray): A complex128 array of data.
        ind (ndarray): A int32 array of indices.
        ptr (ndarray): A int32 array of indptrs.
        vec (ndarray): A complex128 array for the statevector.

    Returns:
        complex: A complex number representing the expectation value.
    """
    nrows = vec.shape[0]
    expt = 0
    for row in prange(nrows):  # pylint: disable=not-an-iterable
        cval = np.conj(vec[row])
        temp = 0
        for jj in range(ptr[row], ptr[row+1]):
            temp += data[jj]*vec[ind[jj]]
        expt += cval*temp
    return expt


def sparse_pauli(num_qubits, idx, kind):
    """Returns a sparse CSR pauli matrix.

    Parameters:
        num_qubits (int): The number of qubits in the statevector.
        idx (int): The index (qubit) on which the Pauli acts.
        kind (str): The kind of Pauli, 'X', 'Y', or 'Z'.

    Returns:
        csr_matrix: A Pauli operator as a sparse CSR matrix.

    Raises:
        ValueError: Invalid 'kind' given.
    """
    if kind == 'X':
        x = 1
        z = 0
    elif kind == 'Y':
        x = 1
        z = 1
    elif kind == 'Z':
        x = 0
        z = 1
    else:
        raise ValueError("kind must be 'X', 'Y', or 'Z'.")
    n = 2**num_qubits
    xs = 2**idx if x else 0
    zs = 2**idx if z else 0
    rows = np.arange(n+1, dtype=np.uint32)
    columns = rows ^ xs
    global_factor = (-1j)**(x*z)
    data = global_factor*(-1)**np.mod(count_set_bits(zs & rows), 2)
    return sp.csr_matrix((data, columns, rows), shape=(n, n))


def bloch_components(rho):
    """Computes the Bloch components of a given statevector or density matrix.

    Parameters:
        rho (ndarray): Input statevector (1D) or density matrix (2D) array.

    Returns:
        list: List of [x,y,z] Bloch components for each qubit in the system.

    Raises:
        ValueError: Invalid input state.
    """
    dim_error = ValueError('Invalid input state.')
    num = np.log2(rho.shape[0])
    if num % 1:
        raise dim_error
    if len(rho.shape) == 2:
        if np.log2(rho.shape[1]) % 1:
            raise dim_error
    num = int(num)
    dims = len(rho.shape)

    out = []
    for i in range(num):
        pauli_singles = [
            sparse_pauli(num, i, 'X'),
            sparse_pauli(num, i, 'Y'),
            sparse_pauli(num, i, 'Z')
        ]
        if dims == 1:
            out.append(list(
                map(lambda x: np.real(expect_psi_csr(x.data, x.indices, x.indptr, rho)),
                    pauli_singles)))

        else:
            out.append(list(
                map(lambda x: np.real(np.trace(x.dot(rho))),
                    pauli_singles)))

    return out



# -*- coding: utf-8 -*-

# This code is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=unused-import

"""Tests for Bloch routines"""

from qiskit import IBMQ, QuantumCircuit
import kaleidoscope.qiskit

PROVIDER = IBMQ.load_account()
backend = PROVIDER.backends.ibmq_vigo


def test_target_backend_attach():
    """Tests attaching target backend"""

    qc = QuantumCircuit(5, 5) >> backend
    qc.h(0)
    qc.cx(0, range(1, 5))
    qc.measure(range(5), range(5))

    assert qc.target_backend == backend


def test_target_backend_transpile():
    """Tests target backend attached to transpiled circuit"""

    qc = QuantumCircuit(5, 5) >> backend
    qc.h(0)
    qc.cx(0, range(1, 5))
    qc.measure(range(5), range(5))

    new_qc = qc.transpile()

    assert new_qc.target_backend == backend



# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
# pylint: disable=wrong-import-position

"""Kaleidoscope qiskit tests module"""



# -*- coding: utf-8 -*-

# This code is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=unnecessary-list-index-lookup, consider-using-generator

"""Interactive Qsphere"""

import numpy as np
import scipy.linalg as la
import scipy.special as spsp
import matplotlib as mpl
import colorcet as cc
from plotly.subplots import make_subplots
import plotly.graph_objects as go
from kaleidoscope.errors import KaleidoscopeError
from .plotly_wrapper import PlotlyWidget, PlotlyFigure
from .bloch.primitives import BSPHERE


def qsphere(state, state_labels=True,
            state_labels_kind='bits',
            as_widget=False):
    """Plots a statevector of qubits using the qsphere
    representation.

    Parameters:
        state (ndarray): Statevector as 1D NumPy array.
        state_labels (bool): Show state labels.
        state_labels_kind (str): 'bits' (default) or 'ints'.
        as_widget (bool): Return a widget instance.

    Returns:
        PlotlyFigure or PlotlyWidget: Figure instance.

    Raises:
        KaleidoscopeError: Invalid statevector input.

    Example:

        .. jupyter-execute::

            from qiskit import QuantumCircuit
            from qiskit.quantum_info import Statevector
            import kaleidoscope.qiskit
            from kaleidoscope.interactive import qsphere

            qc = QuantumCircuit(3)
            qc.h(range(3))
            qc.ch(0,1)
            qc.s(2)
            qc.cz(2,1)
            state = qc.statevector()

            qsphere(state)
    """
    if state.__class__.__name__ in ['Statevector'] \
            and 'qiskit' in state.__class__.__module__:
        state = state.data

    if state.__class__.__name__ in ['DensityMatrix'] \
            and 'qiskit' in state.__class__.__module__:

        if not abs(1-state.data.dot(state.data).trace().real) < 1e-6:
            raise KaleidoscopeError('Input density matrix is not a pure state.')
        # pylint: disable=unexpected-keyword-arg
        _, evecs = la.eig(state.data)
        state = evecs[0].ravel()

    if len(state.shape) == 2:
        if not abs(1-state.dot(state).trace().real) < 1e-6:
            raise KaleidoscopeError('Input density matrix is not a pure state.')
        # pylint: disable=unexpected-keyword-arg
        _, evecs = la.eig(state.data)
        state = evecs[0].ravel()

    if len(state.shape) != 1:
        raise KaleidoscopeError('Input state is not 1D array.')

    if np.log2(state.shape[0]) % 1:
        raise KaleidoscopeError('Input is not a valid statevector of qubits.')

    eps = 1e-8
    norm = mpl.colors.Normalize(vmin=0, vmax=2*np.pi)
    cmap = cc.cm.CET_C1s
    num_qubits = int(np.log2(state.shape[0]))

    xvals = []
    yvals = []
    zvals = []
    colors = []
    bases = []
    probs = []
    marker_sizes = []

    for idx in range(2**num_qubits):
        prob = (state[idx]*state[idx].conj()).real
        if prob > eps:
            elem = bin(idx)[2:].zfill(num_qubits)
            weight = elem.count("1")
            zvalue = -2 * weight / num_qubits + 1
            number_of_divisions = spsp.comb(num_qubits, weight)
            weight_order = _bit_string_index(elem)
            angle = (float(weight) / num_qubits) * (np.pi * 2) + \
                    (weight_order * 2 * (np.pi / number_of_divisions))

            if (weight > num_qubits / 2) or (((weight == num_qubits / 2) and
                                              (weight_order >= number_of_divisions / 2))):
                angle = np.pi - angle - (2 * np.pi / number_of_divisions)

            xvalue = np.sqrt(1 - zvalue ** 2) * np.cos(angle)
            yvalue = np.sqrt(1 - zvalue ** 2) * np.sin(angle)

            bases.append(elem)
            probs.append(prob)
            xvals.append(xvalue)
            yvals.append(yvalue)
            zvals.append(zvalue)

            phase = np.arctan2(state[idx].imag, state[idx].real)
            phase = phase if phase >= 0 else phase+2*np.pi
            colors.append(mpl.colors.rgb2hex(cmap(norm(phase))))
            marker_sizes.append(np.sqrt(prob) * 40)

    if state_labels_kind == 'ints':
        bases = [int(kk, 2) for kk in bases]

    # Output figure instance
    fig = make_subplots(rows=5, cols=5,
                        specs=[[{"type": "scene", "rowspan": 5, "colspan": 5},
                                None, None, None, None],
                               [None, None, None, None, None],
                               [None, None, None, None, None],
                               [None, None, None, None, None],
                               [None, None, None, None,
                                {"rowspan": 1, "colspan": 1, "type": "domain"}
                                ]
                               ]
                        )

    figsize = (350, 350)

    # List for vector annotations, if any
    fig_annotations = []

    fig.add_trace(BSPHERE(), row=1, col=1)

    # latitudes
    for kk in _qsphere_latitudes(zvals):
        fig.add_trace(kk, row=1, col=1)

    fig.add_trace(go.Scatter3d(x=[0], y=[0], z=[0],
                               mode='markers',
                               opacity=0.6,
                               marker=dict(size=4,
                                           color='#555555'),
                               ),
                  row=1, col=1)

    for kk, _ in enumerate(xvals):
        fig.add_trace(go.Scatter3d(x=[0, xvals[kk]], y=[0, yvals[kk]], z=[0, zvals[kk]],
                                   mode="lines",
                                   hoverinfo=None,
                                   opacity=0.5,
                                   line=dict(color=colors[kk], width=3)
                                   ),
                      row=1, col=1
                      )

        if state_labels:
            xanc = 'center'
            if xvals[kk] != 0:
                if xvals[kk] < 0:
                    xanc = 'right'
                else:
                    pass

            yanc = 'middle'
            if zvals[kk] != 0:
                if zvals[kk] < 0:
                    yanc = 'top'
                else:
                    yanc = 'bottom'

            fig_annotations.append(dict(showarrow=False,
                                        x=xvals[kk]*1.1,
                                        y=yvals[kk]*1.1,
                                        z=zvals[kk]*1.1,
                                        text="<b>|{}\u3009</b>".format(bases[kk]),
                                        align='left',
                                        opacity=0.7,
                                        xanchor=xanc,
                                        yanchor=yanc,
                                        xshift=10,
                                        bgcolor="#ffffff",
                                        font=dict(size=10,
                                                  color="#000000",
                                                  ),
                                        )
                                   )

    fig.add_trace(go.Scatter3d(x=xvals, y=yvals, z=zvals,
                               mode='markers',
                               opacity=1,
                               marker=dict(size=marker_sizes,
                                           color=colors),
                               ),
                  row=1, col=1)

    slices = 128
    labels = ['']*slices
    values = [1]*slices

    phase_colors = [mpl.colors.rgb2hex(cmap(norm(2*np.pi*kk/slices))) for kk in range(slices)]

    fig.add_trace(go.Pie(labels=labels, values=values, hole=.6,
                         showlegend=False,
                         textinfo='none',
                         hoverinfo='none',
                         textposition="outside",
                         rotation=90,
                         textfont_size=12,
                         marker=dict(colors=phase_colors)
                         ),
                  row=5, col=5)

    pie_x = fig.data[-1]['domain']['x']
    pie_y = fig.data[-1]['domain']['y']

    fig['layout'].update(annotations=[
        dict(
            xref='paper',
            yref='paper',
            x=(pie_x[1]-pie_x[0])/2+pie_x[0],
            y=(pie_y[1]-pie_y[0])/2+pie_y[0],
            text='Phase',
            xanchor="center",
            yanchor="middle",
            showarrow=False,
            font=dict(size=9),
        ),
        dict(
            xref='paper',
            yref='paper',
            x=pie_x[0]-0.03,
            y=(pie_y[1]-pie_y[0])/2+pie_y[0],
            text='\U0001D70B',
            xanchor="left",
            yanchor="middle",
            showarrow=False,
            font=dict(size=14),
        ),
        dict(
            xref='paper',
            yref='paper',
            x=pie_x[1]+0.03,
            y=(pie_y[1]-pie_y[0])/2+pie_y[0],
            text='0',
            xanchor="right",
            yanchor="middle",
            showarrow=False,
            font=dict(size=12),
        ),
        dict(
            xref='paper',
            yref='paper',
            x=(pie_x[1]-pie_x[0])/2+pie_x[0],
            y=pie_y[1]+0.05,
            text='\U0001D70B/2',
            xanchor="center",
            yanchor="top",
            showarrow=False,
            font=dict(size=12),
        ),
        dict(
            xref='paper',
            yref='paper',
            x=(pie_x[1]-pie_x[0])/2+pie_x[0],
            y=pie_y[0]-0.05,
            text='3\U0001D70B/2',
            xanchor="center",
            yanchor="bottom",
            showarrow=False,
            font=dict(size=12),
        )
    ])

    fig.update_layout(width=figsize[0],
                      height=figsize[1],
                      autosize=False,
                      hoverdistance=50,
                      showlegend=False,
                      scene_aspectmode='cube',
                      margin=dict(r=15, b=15, l=15, t=15),
                      scene=dict(annotations=fig_annotations,
                                 xaxis=dict(showbackground=False,
                                            range=[-1.2, 1.2],
                                            showspikes=False,
                                            visible=False),
                                 yaxis=dict(showbackground=False,
                                            range=[-1.2, 1.2],
                                            showspikes=False,
                                            visible=False),
                                 zaxis=dict(showbackground=False,
                                            range=[-1.2, 1.2],
                                            showspikes=False,
                                            visible=False)),
                      scene_camera=dict(eye=dict(x=0,
                                                 y=-1.4,
                                                 z=0.3)
                                        )
                      )

    if as_widget:
        return PlotlyWidget(fig)

    return PlotlyFigure(fig, modebar=True)


def _lex_index(n, k, lst):
    """Return  the lex index of a combination..

    Args:
        n (int): the total number of options .
        k (int): The number of elements.
        lst (list): list

    Returns:
        int: returns int index for lex order

    Raises:
        KaleidoscopeError: if length of list is not equal to k
    """
    if len(lst) != k:
        raise KaleidoscopeError("list should have length k")
    comb = list(map(lambda x: n - 1 - x, lst))
    dualm = sum([spsp.comb(comb[k - 1 - i], i + 1) for i in range(k)])
    return int(dualm)


def _bit_string_index(s):
    """Return the index of a string of 0s and 1s.

    Parameters:
        s (str): Bitstring.

    Returns:
        int: Index.

    Raises:
        KaleidoscopeError: If string is not binary.
    """
    n = len(s)
    k = s.count("1")
    if s.count("0") != n - k:
        raise KaleidoscopeError("s must be a string of 0 and 1")
    ones = [pos for pos, char in enumerate(s) if char == "1"]
    return _lex_index(n, k, ones)


def _qsphere_latitudes(zvals):
    """Latitude lines for sphere.

    Parameters:
        zvals (int): Input zvals

    Returns:
        list: List of Plotly traces.
    """
    lats = []
    u = np.linspace(0, 2*np.pi, 100)

    if 0 not in zvals:
        zvals = [0] + zvals
    for zv in zvals:
        th = np.arctan2(np.sqrt(1 - zv ** 2), zv)
        xvals = np.sin(th)*np.cos(u)
        yvals = np.sin(th)*np.sin(u)
        zvals = zv*np.ones_like(u)
        lats.append(go.Scatter3d(
            x=xvals, y=yvals, z=zvals,
            mode="lines",
            hoverinfo='skip',
            line=dict(
                color='#1e1e1e' if th == np.pi/2 else '#373737',
                width=2 if th == np.pi/2 else 1
            )
        ))
    return lats



# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Interactive gate map for IBM Quantum Experience devices."""

import plotly.graph_objects as go
from qiskit.providers.ibmq.ibmqbackend import IBMQBackend
from qiskit.providers.fake_provider import FakeBackend
from qiskit.providers.models.backendproperties import BackendProperties
from kaleidoscope.errors import KaleidoscopeError
from kaleidoscope.qiskit.backends.pseudobackend import properties_to_pseudobackend
from kaleidoscope.interactive.plotly_wrapper import PlotlyWidget, PlotlyFigure
from kaleidoscope.qiskit.backends.device_layouts import LAYOUTS


def system_gate_map(
        backend,
        figsize=(None, None),
        label_qubits=True,
        qubit_size=None,
        line_width=None,
        font_size=None,
        qubit_colors="#2f4b7c",
        qubit_labels=None,
        line_colors="#2f4b7c",
        font_color="white",
        background_color='white',
        as_widget=False):
    """Plots an interactive gate map of a device.

    Args:
        backend (IBMQBackend or FakeBackend or DeviceSimulator or Properties): Plot the error map
                                                                               for a backend.
        figsize (tuple): Output figure size (wxh) in pixels.
        label_qubits (bool): Labels for the qubits.
        qubit_size (float): Size of qubit marker.
        line_width (float): Width of lines.
        font_size (float): Font size of qubit labels.
        qubit_colors (str or list): A list of colors for the qubits. If a single color is given,
                                    it's used for all qubits.
        qubit_labels (list): A list of qubit labels
        line_colors (str or list): A list of colors for each line from the coupling map. If a
                                   single color is given, it's used for all lines.
        font_color (str): The font color for the qubit labels.
        background_color (str): The background color, either 'white' or 'black'.
        as_widget (bool): Return the figure as a widget.

    Returns:
        PlotlyFigure or PlotlyWidget: Returned figure instance.

    Raises:
        KaleidoscopeError: Invalid input object.

    Example:
        .. jupyter-execute::

           from qiskit import *
           from kaleidoscope.qiskit.backends import system_gate_map

           pro = IBMQ.load_account()
           backend = pro.backends.ibmq_vigo
           system_gate_map(backend)
    """
    if not isinstance(backend, (IBMQBackend, FakeBackend, BackendProperties)):
        raise KaleidoscopeError('Input is not a valid backend or properties object.')

    if isinstance(backend, BackendProperties):
        backend = properties_to_pseudobackend(backend)

    config = backend.configuration()
    n_qubits = config.n_qubits
    cmap = config.coupling_map

    # set coloring
    if isinstance(qubit_colors, str):
        qubit_colors = [qubit_colors] * n_qubits
    if isinstance(line_colors, str):
        line_colors = [line_colors] * len(cmap) if cmap else []

    if str(n_qubits) in LAYOUTS['layouts'].keys():
        kind = 'generic'
        if backend.name() in LAYOUTS['special_names']:
            if LAYOUTS['special_names'][backend.name()] in LAYOUTS['layouts'][str(n_qubits)]:
                kind = LAYOUTS['special_names'][backend.name()]
        grid_data = LAYOUTS['layouts'][str(n_qubits)][kind]
    else:
        fig = go.Figure()
        fig.update_layout(showlegend=False,
                          plot_bgcolor=background_color,
                          paper_bgcolor=background_color,
                          width=figsize[0], height=figsize[1],
                          margin=dict(t=30, l=0, r=0, b=0))

        if as_widget:
            return PlotlyWidget(fig)
        return PlotlyFigure(fig)

    offset = 0
    if cmap:
        if n_qubits in [14, 15, 16]:
            offset = 1
            if qubit_size is None:
                qubit_size = 24
            if font_size is None:
                font_size = 10
            if line_width is None:
                line_width = 4
            if figsize == (None, None):
                figsize = (400, 200)
        elif n_qubits == 27:
            if qubit_size is None:
                qubit_size = 24
            if font_size is None:
                font_size = 10
            if line_width is None:
                line_width = 4
            if figsize == (None, None):
                figsize = (400, 300)
        else:
            if qubit_size is None:
                qubit_size = 32
            if font_size is None:
                font_size = 14
            if line_width is None:
                line_width = 6
            if figsize == (None, None):
                figsize = (300, 300)
    else:
        if figsize == (None, None):
            figsize = (300, 300)
        if qubit_size is None:
            qubit_size = 30

    fig = go.Figure()

    # Add lines for couplings
    if cmap:
        for ind, edge in enumerate(cmap):
            is_symmetric = False
            if edge[::-1] in cmap:
                is_symmetric = True
            y_start = grid_data[edge[0]][0] + offset
            x_start = grid_data[edge[0]][1]
            y_end = grid_data[edge[1]][0] + offset
            x_end = grid_data[edge[1]][1]

            if is_symmetric:
                if y_start == y_end:
                    x_end = (x_end - x_start) / 2 + x_start
                    x_mid = x_end
                    y_mid = y_start

                elif x_start == x_end:
                    y_end = (y_end - y_start) / 2 + y_start
                    x_mid = x_start
                    y_mid = y_end

                else:
                    x_end = (x_end - x_start) / 2 + x_start
                    y_end = (y_end - y_start) / 2 + y_start
                    x_mid = x_end
                    y_mid = y_end
            else:
                if y_start == y_end:
                    x_mid = (x_end - x_start) / 2 + x_start
                    y_mid = y_end

                elif x_start == x_end:
                    x_mid = x_end
                    y_mid = (y_end - y_start) / 2 + y_start

                else:
                    x_mid = (x_end - x_start) / 2 + x_start
                    y_mid = (y_end - y_start) / 2 + y_start

            fig.add_trace(
                go.Scatter(x=[x_start, x_mid, x_end],
                           y=[-y_start, -y_mid, -y_end],
                           mode="lines",
                           hoverinfo='none',
                           line=dict(width=line_width,
                                     color=line_colors[ind])))

    # Add the qubits themselves
    qubit_text = []
    qubit_str = "<b>Qubit {}"
    for num in range(n_qubits):
        qubit_text.append(qubit_str.format(qubit_labels[num] if qubit_labels else num))

    if qubit_labels is None:
        qubit_labels = [str(ii) for ii in range(n_qubits)]

    if n_qubits > 50:
        if qubit_size is None:
            qubit_size = 20
        if font_size is None:
            font_size = 9

    fig.add_trace(go.Scatter(
        x=[d[1] for d in grid_data],
        y=[-d[0]-offset for d in grid_data],
        mode="markers+text",
        marker=go.scatter.Marker(size=qubit_size,
                                 color=qubit_colors,
                                 opacity=1),
        text=qubit_labels if label_qubits else '',
        textposition="middle center",
        textfont=dict(size=font_size, color=font_color),
        hoverinfo="text" if label_qubits else 'none',
        hovertext=qubit_text))

    fig.update_xaxes(visible=False)
    _range = None
    if offset:
        _range = [-3.5, 0.5]
    fig.update_yaxes(visible=False, range=_range)

    fig.update_layout(showlegend=False,
                      plot_bgcolor=background_color,
                      paper_bgcolor=background_color,
                      width=figsize[0], height=figsize[1],
                      margin=dict(t=30, l=0, r=0, b=0))

    if as_widget:
        return PlotlyWidget(fig)
    return PlotlyFigure(fig)



# -*- coding: utf-8 -*-

# This code is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# The Hamming distance and internal plot histogram data are part of
# Qiskit.  The latter will be modified at some point.

"""Interactive histogram from experiment counts"""

import functools
from collections import Counter, OrderedDict
import numpy as np
import plotly.graph_objects as go
from kaleidoscope.errors import KaleidoscopeError
from kaleidoscope.colors import COLORS1, COLORS2, COLORS3, COLORS4, COLORS5, COLORS14
from kaleidoscope.colors.utils import find_text_color
from .plotly_wrapper import PlotlyWidget, PlotlyFigure


def hamming_distance(str1, str2):
    """Calculate the Hamming distance between two bit strings

    Args:
        str1 (str): First string.
        str2 (str): Second string.
    Returns:
        int: Distance between strings.
    Raises:
        ValueError: Strings not same length
    """
    if len(str1) != len(str2):
        raise ValueError('Strings not same length.')
    return sum(s1 != s2 for s1, s2 in zip(str1, str2))


VALID_SORTS = ['asc', 'desc', 'hamming']
DIST_MEAS = {'hamming': hamming_distance}


def probability_distribution(data, figsize=(None, None), colors=None,
                             scale='linear',
                             number_to_keep=None,
                             sort='asc', target_string=None,
                             legend=None, bar_labels=True,
                             state_labels_kind='bits',
                             state_labels=None,
                             title=None,
                             plot_background_color='#e5e0df',
                             background_color='#ffffff',
                             as_widget=False):
    """Interactive histogram plot of probability distributions.

    Parameters:
        data (list or dict): This is either a list of dictionaries or a single
            dict containing the values to represent (ex {'001': 130})
        figsize (tuple): Figure size in pixels.
        colors (list or str): String or list of strings for histogram bar colors.
        scale (str): Probability scale 'linear' (default) or 'log'.
        number_to_keep (int): The number of terms to plot and rest
            is made into a single bar called 'rest'.
        sort (string): Could be 'asc', 'desc', or 'hamming'.
        target_string (str): Target string if 'sort' is a distance measure.
        legend(list): A list of strings to use for labels of the data.
            The number of entries must match the length of data (if data is a
            list or 1 if it's a dict)
        bar_labels (bool): Label each bar in histogram with probability value.
        state_labels_kind (str): 'bits' (default) or 'ints'.
        state_labels (list): A list of custom state labels.
        title (str): A string to use for the plot title.
        plot_background_color (str): Set the background color behind histogram bars.
        background_color (str): Set the background color for axes.
        as_widget (bool): Return figure as an ipywidget.

    Returns:
        PlotlyFigure or PlotlyWidget:
            A figure for the rendered histogram.

    Raises:
        ValueError: When legend is provided and the length doesn't
                    match the input data.
        ImportError: When failed to load plotly.
        KaleidoscopeError: When state labels is not correct length.

    Example:
        .. jupyter-execute::

            from qiskit import *
            import kaleidoscope.qiskit
            from kaleidoscope.qiskit.services import Simulators
            from kaleidoscope.interactive import probability_distribution

            sim = Simulators.aer_vigo_simulator

            qc = QuantumCircuit(3, 3) >> sim
            qc.h(1)
            qc.cx(1,0)
            qc.cx(1,2)
            qc.measure(range(3), range(3))

            counts = qc.transpile().sample().result_when_done()
            probability_distribution(counts)
    """

    if sort not in VALID_SORTS:
        raise ValueError("Value of sort option, %s, isn't a "
                         "valid choice. Must be 'asc', "
                         "'desc', or 'hamming'")
    if sort in DIST_MEAS and target_string is None:
        err_msg = 'Must define target_string when using distance measure.'
        raise ValueError(err_msg)

    if isinstance(data, dict):
        data = [data]

    if legend and len(legend) != len(data):
        raise ValueError("Length of legendL (%s) doesn't match "
                         "number of input executions: %s" %
                         (len(legend), len(data)))

    text_color = find_text_color(background_color)

    labels = list(sorted(
        functools.reduce(lambda x, y: x.union(y.keys()), data, set())))
    if number_to_keep is not None:
        labels.append('rest')

    if sort in DIST_MEAS:
        dist = []
        for item in labels:
            dist.append(DIST_MEAS[sort](item, target_string))

        labels = [list(x) for x in zip(*sorted(zip(dist, labels),
                                               key=lambda pair: pair[0]))][1]
    # Set bar colors
    if colors is None:
        if len(data) == 1:
            colors = COLORS1
        elif len(data) == 2:
            colors = COLORS2
        elif len(data) == 3:
            colors = COLORS3
        elif len(data) == 4:
            colors = COLORS4
        elif len(data) == 5:
            colors = COLORS5
        else:
            colors = COLORS14
    elif isinstance(colors, str):
        colors = [colors]

    width = 1/(len(data)+1)  # the width of the bars

    labels_dict, all_pvalues, _ = _plot_histogram_data(data,
                                                       labels,
                                                       number_to_keep)

    fig = go.Figure()
    for item, _ in enumerate(data):
        xvals = []
        yvals = []
        for idx, val in enumerate(all_pvalues[item]):
            xvals.append(idx+item*width)
            yvals.append(val)

        labels = list(labels_dict.keys())
        if state_labels_kind == 'ints':
            labels = [int(label, 2) for label in labels]
        if state_labels:
            if len(state_labels) != len(labels):
                raise KaleidoscopeError('Number of input state labels does not match data.')
            labels = state_labels

        hover_template = "<b>{x}</b><br>P = {y}"
        hover_text = [hover_template.format(x=labels[kk],
                                            y=np.round(yvals[kk], 3)) for kk in range(len(yvals))]

        fig.add_trace(go.Bar(x=xvals,
                             y=yvals,
                             width=width,
                             hoverinfo="text",
                             hovertext=hover_text,
                             marker_color=colors[item % len(colors)],
                             name=legend[item] if legend else '',
                             text=np.round(yvals, 3) if bar_labels else None,
                             textposition='auto'
                             ))

    xaxes_labels = list(labels_dict.keys())
    if state_labels_kind == 'ints':
        xaxes_labels = [int(label, 2) for label in xaxes_labels]

    if state_labels:
        if len(state_labels) != len(xaxes_labels):
            raise KaleidoscopeError('Number of input state labels does not match data.')
        xaxes_labels = state_labels

    fig.update_xaxes(title='Basis state',
                     titlefont_size=16,
                     tickvals=list(range(len(labels_dict.keys()))),
                     ticktext=xaxes_labels,
                     tickfont_size=14,
                     showline=True, linewidth=1,
                     linecolor=text_color if text_color == 'white' else None,
                     )

    fig.update_yaxes(title='Probability',
                     titlefont_size=16,
                     tickfont_size=14,
                     showline=True, linewidth=1,
                     linecolor=text_color if text_color == 'white' else None,
                     )

    if scale == 'log':
        lower = np.min([min(item.values())/sum(item.values()) for item in data])
        lower = int(np.floor(np.log10(lower)))
        fig.update_yaxes(type="log", range=[lower, 0])
        fig.update_layout(yaxis=dict(tickmode='array',
                          tickvals=[10**k for k in range(lower, 1)],
                          ticktext=["10<sup>{}</sup>".format(k) for k in range(lower, 1)]
                          ))

    fig.update_layout(xaxis_tickangle=-70,
                      showlegend=(legend is not None),
                      width=figsize[0],
                      height=figsize[1],
                      plot_bgcolor=plot_background_color,
                      paper_bgcolor=background_color,
                      title=dict(text=title, x=0.5),
                      title_font_size=18,
                      font=dict(color=text_color),
                      )
    if as_widget:
        return PlotlyWidget(fig)

    return PlotlyFigure(fig)


def _plot_histogram_data(data, labels, number_to_keep):
    """Generate the data needed for plotting counts.

    Parameters:
        data (list or dict): This is either a list of dictionaries or a single
            dict containing the values to represent (ex {'001': 130})
        labels (list): The list of bitstring labels for the plot.
        number_to_keep (int): The number of terms to plot and rest
            is made into a single bar called 'rest'.

    Returns:
        tuple: tuple containing:
            (dict): The labels actually used in the plotting.
            (list): List of ndarrays for the bars in each experiment.
            (list): Indices for the locations of the bars for each
                    experiment.
    """
    labels_dict = OrderedDict()

    all_pvalues = []
    all_inds = []
    for execution in data:
        if number_to_keep is not None:
            data_temp = dict(Counter(execution).most_common(number_to_keep))
            data_temp["rest"] = sum(execution.values()) - sum(data_temp.values())
            execution = data_temp
        values = []
        for key in labels:
            if key not in execution:
                if number_to_keep is None:
                    labels_dict[key] = 1
                    values.append(0)
                else:
                    values.append(-1)
            else:
                labels_dict[key] = 1
                values.append(execution[key])
        values = np.array(values, dtype=float)
        where_idx = np.where(values >= 0)[0]
        pvalues = values[where_idx] / sum(values[where_idx])

        all_pvalues.append(pvalues)
        numelem = len(values[where_idx])
        ind = np.arange(numelem)  # the x locations for the groups
        all_inds.append(ind)

    return labels_dict, all_pvalues, all_inds



# -*- coding: utf-8 -*-

# This code is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for Bloch routines"""

import numpy as np
from kaleidoscope.interactive.bloch.utils import bloch_components


def test_bloch_components():
    """Tests the Bloch components function"""

    # |0> state
    state = np.array([1, 0], dtype=complex)
    comp = bloch_components(state)[0]
    assert np.allclose(comp, [0, 0, 1])

    # |1> state
    state = np.array([0, 1], dtype=complex)
    comp = bloch_components(state)[0]
    assert np.allclose(comp, [0, 0, -1])

    # (|00> + |11>)/sqrt(2)
    state = np.array([1/np.sqrt(2), 0, 0, 1/np.sqrt(2)], dtype=complex)
    comp = bloch_components(state)
    assert np.allclose(comp[0], [0.0, 0.0, 0.0])
    assert np.allclose(comp[1], [0.0, 0.0, 0.0])



# -*- coding: utf-8 -*-
# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
===================================================
Interactive Plots (:mod:`kaleidoscope.interactive`)
===================================================

.. currentmodule:: kaleidoscope.interactive

Distributions
=============
.. autosummary::
   :toctree: ../stubs/

   probability_distribution


Statevectors
============
.. autosummary::
   :toctree: ../stubs/

   bloch_sphere
   bloch_disc
   bloch_multi_disc
   qsphere


Plotly wrappers
===============

.. autosummary::
   :toctree: ../stubs/

   PlotlyFigure
   PlotlyWidget
"""

from .histogram import probability_distribution
from .bloch.bloch2d import bloch_disc, bloch_multi_disc
from .bloch.bloch3d import bloch_sphere
from .qsphere import qsphere
from .plotly_wrapper import PlotlyFigure, PlotlyWidget



# -*- coding: utf-8 -*-

# This code is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.

"""A module of Plotly class wrappers.
"""

import plotly.graph_objects as go


class PlotlyFigure():
    """A simple wrapper around Plotly Figure class.

    Allows the figures to be more or less drop in replacements
    for Matplotlib Figures, e.g. savefig is a method here.
    """
    def __init__(self, fig, modebar=False):
        self._fig = fig
        self._modebar = modebar

    def __repr__(self):
        return self._fig.__repr__()

    def _ipython_display_(self):
        import plotly.io as pio

        if pio.renderers.render_on_display and pio.renderers.default:
            pio.show(self._fig, config={'displayModeBar': self._modebar,
                                        'editable': False})
        else:
            print(repr(self))

    def show(self, *args, **kwargs):
        """Display the figure.
        """
        import plotly.io as pio

        config = {}
        if 'config' not in kwargs:
            config = {'displayModeBar': self._modebar,
                      'editable': False}

        return pio.show(self._fig, *args, config=config, **kwargs)

    def savefig(self, filename, figsize=(None, None), scale=1, transparent=False):
        """Save the figure.

        Parameters:
            filename (str): Filename to save to.
            figsize (tuple): Figure size (W x H) in pixels.
            scale (float): Scales the output figure.
            transparent (bool): Transparent background.
        """
        if transparent:
            plot_color = self._fig.layout['plot_bgcolor']
            paper_color = self._fig.layout['paper_bgcolor']
            self._fig.update_layout(paper_bgcolor='rgba(0,0,0,0)',
                                    plot_bgcolor='rgba(0,0,0,0)')
        self._fig.write_image(filename, width=figsize[0], height=figsize[1], scale=scale)
        if transparent:
            self._fig.update_layout(plot_bgcolor=plot_color,
                                    paper_bgcolor=paper_color)


class PlotlyWidget(go.FigureWidget):
    """A wrapper around the Plotly widget class.
    """
    def show(self, *args, **kwargs):
        """Display the figure.
        """
        import plotly.io as pio

        config = {}
        if 'config' not in kwargs:
            config = {'displayModeBar': False,
                      'editable': False}

        return pio.show(self, *args, config=config, **kwargs)

    def savefig(self, filename, figsize=(None, None), scale=1, transparent=False):
        """Safe the figure as a static image.

        Parameters:
            filename (str): Name of the file to which the image is saved.
            figsize (tuple): Size of figure in pixels.
            scale (float): Scale factor for non-vectorized image formats.
            transparent (bool): Set the background to transparent.
        """
        if transparent:
            plot_color = self.layout['plot_bgcolor']
            paper_color = self.layout['paper_bgcolor']
            self.update_layout(paper_bgcolor='rgba(0,0,0,0)',
                               plot_bgcolor='rgba(0,0,0,0)')

        self.write_image(filename, width=figsize[0], height=figsize[1], scale=scale)
        if transparent:
            self.update_layout(plot_bgcolor=plot_color,
                               paper_bgcolor=paper_color)



# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tools for colors"""
import colorcet as cc

# The Blue->Magenta->White sequential color map
BMW = cc.cm.bmw
# The Black->Blue->Cyan sequential color map
KBC = cc.cm.kbc
# The Black->Gray->Yellow diverging color map
BJY = cc.cm.bjy

# Dark2 from colorbrewer
DARK2 = ["#1b9e77", "#d95f02", "#7570b3", "#e7298a",
         "#66a61e", "#e6ab02", "#a6761d", "#666666"]

COLORS14 = ['#6929c4', '#1192e8', '#005d5d', '#9f1853', '#fa4d56',
            '#570408', '#198038', '#002d9c', '#ee538b', '#b28600',
            '#009d9a', '#012749', '#8a3800', '#a56eff']
COLORS5 = ['#6929c4', '#1192e8', '#005d5d', '#9f1853', '#570408']
COLORS4 = ['#6929c4', '#012749', '#009d9a', '#ee538b']
COLORS3 = ['#a56eff', '#005d5d', '#9f1853']
COLORS2 = ['#6929c4', '#009d9a']
COLORS1 = ['#6929c4']



# -*- coding: utf-8 -*-

# This code is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Interactive Bloch sphere"""

import numpy as np
import plotly.graph_objects as go
from kaleidoscope.colors import COLORS14
from kaleidoscope.colors.utils import hex_to_rgb
from kaleidoscope.interactive.plotly_wrapper import PlotlyFigure, PlotlyWidget
from kaleidoscope.interactive.bloch.primitives import (BSPHERE, LATS, LONGS, ZAXIS, YAXIS, XAXIS,
                                                       Z0LABEL, Z1LABEL, YLABEL, XLABEL)
from kaleidoscope.interactive.bloch.utils import bloch_components
from kaleidoscope.errors import KaleidoscopeError


def bloch_sphere(vectors=None,
                 vectors_color=None,
                 vectors_alpha=None,
                 vectors_annotation=False,
                 points=None,
                 points_color=None,
                 points_alpha=None,
                 figsize=(350, 350),
                 label_fontsize=16,
                 annotation_fontsize=10,
                 as_widget=False):
    """Generates a Bloch sphere from a given collection of vector
    and/or points data expressed in cartesian coordinates, [x, y, z].

    Parameters:
        vectors (list, ndarray): Collection of one or more vectors to display.
        vectors_color (str or list): List of colors to use when plotting vectors.
        vectors_alpha (float or list): List of alphas to use when plotting vectors.
        vectors_annotation (bool or list): Boolean values to determine if a
                                           annotation should be displayed.
        points (list, ndarray): Collection of one or more points to display.
        points_color (str or list): List of colors to use when plotting points.
        points_alpha (float or list): List of alphas to use when plotting points.
        figsize (tuple): Figure size in pixels.
        label_fontsize (int): Font size for axes labels.
        annotation_fontsize (int): Font size for annotations.
        as_widget (bool): Return plot as a widget.

    Returns:
        PlotlyFigure or  PlotlyWidget: A Plotly figure or widget instance

    Raises:
        ValueError: Input lengths do not match.
        KaleidoscopeError: Invalid vector input.

    Example:
        .. jupyter-execute::

            import numpy as np
            from matplotlib.colors import LinearSegmentedColormap, rgb2hex
            from kaleidoscope.interactive import bloch_sphere

            cm = LinearSegmentedColormap.from_list('graypurple', ["#999999", "#AA00FF"])

            pointsx = [[0, -np.sin(th), np.cos(th)] for th in np.linspace(0, np.pi/2, 20)]
            pointsz = [[np.sin(th), -np.cos(th), 0] for th in np.linspace(0, 3*np.pi/4, 30)]

            points = pointsx + pointsz

            points_alpha = [np.linspace(0.8, 1, len(points))]

            points_color = [[rgb2hex(cm(kk)) for kk in np.linspace(-1,1,len(points))]]

            vectors_color = ["#777777", "#AA00FF"]

            bloch_sphere(points=points,
                         vectors=[[0, 0, 1], [1/np.sqrt(2), 1/np.sqrt(2), 0]],
                         vectors_color=vectors_color,
                         points_alpha=points_alpha,
                         points_color=points_color)
    """

    # Output figure instance
    fig = go.Figure()

    # List for vector annotations, if any
    fig_annotations = []

    idx = 0
    if points is not None:

        nest_depth = nest_level(points)
        # Take care of single point passed
        if nest_depth == 1:
            points = [[points]]
        # A single list of points passes
        elif nest_depth == 2:
            points = [points]
        # nest_depth = 3 means multiple lists passed

        if points_color is None:
            # passed a single point
            if nest_depth == 1:
                points_color = [COLORS14[0]]
            elif nest_depth == 2:
                points_color = [[COLORS14[kk % 14] for kk in range(len(points[0]))]]

            elif nest_depth == 3:
                points_color = []
                for kk, pnts in enumerate(points):
                    points_color.append(COLORS14[kk % 14]*len(pnts))

        if nest_depth == 2 and nest_level(points_color) == 1:
            points_color = [points_color]

        if isinstance(points_color, str):
            points_color = [points_color]

        if points_alpha is None:
            points_alpha = [[1.0]*len(p) for p in points]

        if nest_depth == 2 and nest_level(points_alpha) == 1:
            points_alpha = [points_alpha]

        if isinstance(points_alpha, (int, float)):
            points_alpha = [[points_alpha]]

        for idx, point_collection in enumerate(points):
            x_pnts = []
            y_pnts = []
            z_pnts = []
            if isinstance(points_color[idx], str):
                _colors = [points_color[idx]]*len(point_collection)
            else:
                _colors = points_color[idx]

            if len(points_alpha[idx]) != len(point_collection):
                err_str = 'number of alpha values ({}) does not equal number of points ({})'
                raise ValueError(err_str.format(len(points_alpha[idx]), len(x_pnts)))

            mcolors = []
            for kk, point in enumerate(point_collection):
                x_pnts.append(point[0])
                y_pnts.append(point[1])
                z_pnts.append(point[2])

                mcolors.append("rgba({},{},{},{})".format(*hex_to_rgb(_colors[kk]),
                                                          points_alpha[idx][kk]))

            fig.add_trace(go.Scatter3d(x=x_pnts, y=y_pnts, z=z_pnts,
                                       mode='markers',
                                       marker=dict(size=7, color=mcolors),
                                       )
                          )
            idx += 1

    if vectors is not None:

        if vectors.__class__.__name__ in ['Statevector'] \
                and 'qiskit' in vectors.__class__.__module__:
            vectors = bloch_components(vectors.data)
        elif vectors.__class__.__name__ in ['DensityMatrix'] \
                and 'qiskit' in vectors.__class__.__module__:
            vectors = bloch_components(vectors.data)

        elif not isinstance(vectors[0], (list, np.ndarray)):
            if vectors[0].__class__.__name__ not in ['Statevector']:
                vectors = [[vectors[0], vectors[1], vectors[2]]]

        new_vecs = []
        for vec in vectors:
            if vec.__class__.__name__ in ['Statevector'] and 'qiskit' in vec.__class__.__module__:
                # pylint: disable=no-member
                new_vecs.append(bloch_components(vec.data)[0])
            else:
                nst_lvl = nest_level(vec)
                if nst_lvl == 1:
                    new_vecs.append(vec)
                elif nst_lvl == 2:
                    new_vecs.append(vec[0])
                else:
                    raise KaleidoscopeError("Invalid vector input.")

        if vectors_color is None:
            vectors_color = [COLORS14[kk+idx % 14] for kk in range(len(new_vecs))]

        if isinstance(vectors_color, str):
            vectors_color = [vectors_color]

        if vectors_alpha is None:
            vectors_alpha = [1.0]*len(new_vecs)

        if isinstance(vectors_alpha, (int, float)):
            vectors_alpha = [vectors_alpha]

        if vectors_annotation is True:
            vectors_annotation = [True]*len(new_vecs)
        elif not vectors_annotation:
            vectors_annotation = [False]*len(new_vecs)

        eps = 1e-12

        for idx, vec in enumerate(new_vecs):
            vec = np.asarray(vec)
            if np.linalg.norm(vec) > 1.0 + eps:
                raise ValueError('Vector norm must be <= 1.')
            # So that line does not go out of arrow head
            vec_line = vec / 1.05

            color_str = "rgba({},{},{},{})".format(*hex_to_rgb(vectors_color[idx]),
                                                   vectors_alpha[idx])

            fig.add_trace(go.Scatter3d(x=[0, vec_line[0]], y=[0, vec_line[1]], z=[0, vec_line[2]],
                                       mode="lines",
                                       hoverinfo=None,
                                       line=dict(color=color_str, width=10)
                                       )
                          )

            fig.add_trace(go.Cone(x=[vec[0]], y=[vec[1]], z=[vec[2]],
                                  u=[vec[0]], v=[vec[1]], w=[vec[2]],
                                  sizemode="absolute",
                                  showscale=False,
                                  opacity=vectors_alpha[idx],
                                  colorscale=[vectors_color[idx],
                                              vectors_color[idx]],
                                  sizeref=0.25,
                                  anchor="tip")
                          )

            if vectors_annotation[idx]:
                fig_annotations.append(dict(showarrow=False,
                                            x=vec[0]*1.05,
                                            y=vec[1]*1.05,
                                            z=vec[2]*1.05,
                                            text="[{},<br> {},<br> {}]".format(round(vec[0], 3),
                                                                               round(vec[1], 3),
                                                                               round(vec[2], 3)),
                                            align='left',
                                            borderpad=3,
                                            xanchor='right' if vec[1] <= 0 else "left",
                                            xshift=10,
                                            bgcolor="#53565F",
                                            font=dict(size=annotation_fontsize,
                                                      color="#ffffff",
                                                      family="Courier New, monospace",
                                                      ),
                                            )
                                       )
    # Start construction of sphere
    # Sphere
    fig.add_trace(BSPHERE())

    # latitudes
    for kk in LATS:
        fig.add_trace(kk)

    # longitudes
    for kk in LONGS:
        fig.add_trace(kk)

    # z-axis
    fig.add_trace(ZAXIS)
    # x-axis
    fig.add_trace(XAXIS)
    # y-axis
    fig.add_trace(YAXIS)

    # zaxis label
    fig.add_trace(Z0LABEL(fontsize=label_fontsize))
    fig.add_trace(Z1LABEL(fontsize=label_fontsize))
    # xaxis label
    fig.add_trace(XLABEL(fontsize=label_fontsize))
    # yaxis label
    fig.add_trace(YLABEL(fontsize=label_fontsize))

    fig.update_layout(width=figsize[0],
                      height=figsize[1],
                      autosize=False,
                      hoverdistance=50,
                      showlegend=False,
                      scene_aspectmode='cube',
                      margin=dict(r=0, b=0, l=0, t=0),
                      scene=dict(annotations=fig_annotations,
                                 xaxis=dict(showbackground=False,
                                            range=[-1.2, 1.2],
                                            showspikes=False,
                                            visible=False),
                                 yaxis=dict(showbackground=False,
                                            range=[-1.2, 1.2],
                                            showspikes=False,
                                            visible=False),
                                 zaxis=dict(showbackground=False,
                                            range=[-1.2, 1.2],
                                            showspikes=False,
                                            visible=False)),
                      scene_camera=dict(eye=dict(x=1.5,
                                                 y=0.4,
                                                 z=0.4)
                                        )
                      )
    if as_widget:
        return PlotlyWidget(fig)

    return PlotlyFigure(fig, modebar=True)


def nest_level(lst):
    """Determine how much nesting is in a list/ ndarray.

    Parameters:
        lst (list or ndarray): Input array-like object.

    Returns:
        int: Level of nesting.
    """
    if not isinstance(lst, (list, np.ndarray)):
        return 0
    if isinstance(lst, list):
        if not lst:
            return 1
    else:
        if isinstance(lst, np.ndarray):
            if not all(lst):
                return 1
    return max(nest_level(item) for item in lst) + 1



# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=too-many-return-statements

# Modified from Qiskit version for better performance over
# wider range of values.

"""Check if number close to values of PI
"""

import numpy as np

N, D = np.meshgrid(np.arange(1, 17), np.arange(1, 17))
FRAC_MESH = N / D * np.pi


def pi_check(inpt, eps=1e-6, ndigits=5):
    """ Computes if a number is close to an integer
    fraction or multiple of PI and returns the
    corresponding string.

    Args:
        inpt (float): Number to check.
        eps (float): EPS to check against.
        ndigits (int): Number of digits to print
                       if returning raw inpt.

    Returns:
        str: string representation of output.
    """

    if abs(inpt) < 1e-14:
        return '0'
    pi = 'pi'
    val = inpt / np.pi
    if abs(val) >= 1 - eps:
        if abs(abs(val) - abs(round(val))) < eps:
            val = int(round(val))
            if val == 1:
                str_out = '{}'.format(pi)
            elif val == -1:
                str_out = '-{}'.format(pi)
            else:
                str_out = '{}{}'.format(val, pi)
            return str_out

    if abs(abs(val) - abs(round(val))) < eps:
        val = int(round(val))
        if val > 0:
            str_out = '{}/{}'.format(pi, val)
        else:
            str_out = '-{}/{}'.format(pi, abs(val))
        return str_out

    # Look for all fracs in 16
    abs_val = abs(inpt)
    frac = np.where(np.abs(abs_val - FRAC_MESH) < 1e-8)
    if frac[0].shape[0]:
        numer = int(frac[1][0]) + 1
        denom = int(frac[0][0]) + 1
        if inpt < 0:
            numer *= -1

        if numer == 1 and denom == 1:
            str_out = '{}'.format(pi)
        elif numer == -1 and denom == 1:
            str_out = '-{}'.format(pi)
        elif numer == 1:
            str_out = '{}/{}'.format(pi, denom)
        elif numer == -1:
            str_out = '-{}/{}'.format(pi, denom)
        elif denom == 1:
            str_out = '{}/{}'.format(numer, pi)
        else:
            str_out = '{}{}/{}'.format(numer, pi, denom)
        return str_out
    # nothing found
    str_out = '%.{}g'.format(ndigits) % inpt
    return str_out



# -*- coding: utf-8 -*-

# This code is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Primitives for building a 3D sphere"""

import numpy as np
import plotly.graph_objects as go

u = np.linspace(0, 2*np.pi, 100)
v = np.linspace(0, 2*np.pi, 100)
x = np.outer(np.cos(u), np.sin(v))
y = np.outer(np.sin(u), np.sin(v))
z = np.outer(np.ones(np.size(u)), np.cos(v))


def BSPHERE(color="#F1EBEA", opacity=0.15):
    """Surface of sphere.

    Parameters:
        color (str): Hex color for sphere.
        opacity (float): Opacity of sphere.

    Returns:
        surface: Plotly Surface trace.
    """
    surface = go.Surface(x=x, y=y, z=z,
                         opacity=opacity,
                         showscale=False,
                         hoverinfo='skip',
                         colorscale=[[0, color],
                                     [1, color]],
                         contours=dict(
                             x=dict(highlight=False),
                             y=dict(highlight=False),
                             z=dict(highlight=False))
                         )
    return surface


def latitudes(num_lines=7):
    """Latitude lines for sphere.

    Parameters:
        num_lines (int): Number of lines to draw.

    Returns:
        list: List of Plotly traces.
    """
    lats = []
    for th in np.linspace(0, np.pi, num_lines):
        xvals = np.sin(th)*np.cos(u)
        yvals = np.sin(th)*np.sin(u)
        zvals = np.cos(th)*np.ones_like(u)
        lats.append(go.Scatter3d(
            x=xvals, y=yvals, z=zvals,
            mode="lines",
            hoverinfo='skip',
            line=dict(
                color='#1e1e1e' if th == np.pi/2 else '#373737',
                width=2 if th == np.pi/2 else 1
            )
        ))
    return lats


LATS = latitudes()


def longitudes(num_lines=7):
    """Longitude lines for sphere.

    Parameters:
        num_lines (int): Number of lines to draw.

    Returns:
        list: List of Plotly traces.
    """
    longi = []
    for th in np.linspace(0, 2*np.pi, num_lines):
        xvals = np.cos(th)*np.sin(u)*np.ones_like(u)
        yvals = np.sin(th)*np.sin(u)*np.ones_like(u)
        zvals = np.cos(u)
        longi.append(go.Scatter3d(x=xvals, y=yvals, z=zvals,
                                  mode="lines",
                                  hoverinfo='skip',
                                  line=dict(color='#373737',
                                            width=1)
                                  )
                     )
    return longi


LONGS = longitudes()

ZAXIS = go.Scatter3d(x=[0, 0], y=[0, 0], z=[-1, 1],
                     mode="lines",
                     hoverinfo='skip',
                     line=dict(color='#1e1e1e',
                               width=3)
                     )


YAXIS = go.Scatter3d(x=[0, 0], y=[-1, 1], z=[0, 0],
                     mode="lines",
                     hoverinfo='skip',
                     line=dict(color='#1e1e1e',
                               width=3)
                     )

XAXIS = go.Scatter3d(x=[-1, 1], y=[0, 0], z=[0, 0],
                     mode="lines",
                     hoverinfo='skip',
                     line=dict(color='#1e1e1e',
                               width=3)
                     )


def Z0LABEL(fontsize=16):
    """Z0 state label for sphere.

    Parameters:
        fontsize (int): Fontsize for label.

    Returns:
        label: Plotly Scatter3d trace.

    """
    label = go.Scatter3d(x=[0], y=[0], z=[1.1],
                         mode="text",
                         textposition='middle center',
                         hoverinfo='skip',
                         text='\u007C0\u3009',
                         textfont=dict(size=fontsize)
                         )
    return label


def Z1LABEL(fontsize=16):
    """Z1 state label for sphere.

    Parameters:
        fontsize (int): Fontsize for label.

    Returns:
        label: Plotly Scatter3d trace.

    """
    label = go.Scatter3d(x=[0], y=[0], z=[-1],
                         mode="text",
                         textposition='bottom center',
                         hoverinfo='skip',
                         text='\u007C1\u3009',
                         textfont=dict(size=fontsize)
                         )
    return label


def XLABEL(fontsize=16):
    """X+ state label for sphere.

    Parameters:
        fontsize (int): Fontsize for label.

    Returns:
        label: Plotly Scatter3d trace.
    """
    label = go.Scatter3d(x=[1.2], y=[0], z=[0],
                         mode="text",
                         textposition='middle center',
                         hoverinfo='skip',
                         text='\u007C\u002Bx\u3009',
                         textfont=dict(size=fontsize)
                         )
    return label


def YLABEL(fontsize=16):
    """Y+ state label for sphere.

    Parameters:
        fontsize (int): Fontsize for label.

    Returns:
        label: Plotly Scatter3d trace.
    """
    label = go.Scatter3d(x=[0], y=[1.2], z=[0],
                         mode="text",
                         textposition='middle center',
                         hoverinfo='skip',
                         text='\u007C\u002By\u3009',
                         textfont=dict(size=fontsize)
                         )
    return label



# -*- coding: utf-8 -*-

# This code is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for Bloch routines"""

import pytest
from qiskit import QuantumCircuit
from qiskit. quantum_info import DensityMatrix, partial_trace
from kaleidoscope import qsphere
from kaleidoscope.errors import KaleidoscopeError


def test_qsphere_bad_dm_input():
    """Tests the qsphere raises when passed impure dm"""

    qc = QuantumCircuit(3)
    qc.h(0)
    qc.cx(0, 1)
    qc.cx(1, 2)
    dm = DensityMatrix.from_instruction(qc)
    pdm = partial_trace(dm, [0, 1])
    with pytest.raises(KaleidoscopeError):
        assert qsphere(pdm)



# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
==================================================================
IBM Quantum Backend Routines (:mod:`kaleidoscope.qiskit.backends`)
==================================================================

.. currentmodule:: kaleidoscope.qiskit.backends

System visualizations
=====================

.. autosummary::
   :toctree: ../stubs/

   system_error_map
   system_gate_map
   cnot_error_density

"""

from .mpl.cnot_err import cnot_error_density
from .interactive import system_error_map, system_gate_map



# -*- coding: utf-8 -*-

#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""Kaleidoscope

Kaleidoscope is a visualization toolkit for quantum computation.
It also has a Qiskit interface for plotting details about the
IBM roster of Quantum devices.
"""

import os
import sys
import subprocess
import setuptools


MAJOR = 0
MINOR = 2
MICRO = 0

ISRELEASED = False
VERSION = '%d.%d.%d' % (MAJOR, MINOR, MICRO)

REQUIREMENTS = ['numpy>=1.15',
                'scipy>=1.3',
                'numba>=0.46',
                'plotly>=4.6',
                'matplotlib>=3.1',
                'seaborn>=0.9.0',
                'jupyter',
                'colorcet',
               ]

PACKAGES = setuptools.find_namespace_packages()
PACKAGE_DATA = {
}

DOCLINES = __doc__.split('\n')
DESCRIPTION = DOCLINES[0]
LONG_DESCRIPTION = "\n".join(DOCLINES[2:])


def git_short_hash():
    try:
        git_str = "+" + os.popen('git log -1 --format="%h"').read().strip()
    except:  # pylint: disable=bare-except
        git_str = ""
    else:
        if git_str == '+': #fixes setuptools PEP issues with versioning
            git_str = ''
    return git_str

FULLVERSION = VERSION
if not ISRELEASED:
    FULLVERSION += '.dev'+str(MICRO)+git_short_hash()

def write_version_py(filename='kaleidoscope/version.py'):
    cnt = """\
# THIS FILE IS GENERATED FROM KALEIDOSCOPE SETUP.PY
# pylint: disable=missing-module-docstring
short_version = '%(version)s'
version = '%(fullversion)s'
release = %(isrelease)s
"""
    a = open(filename, 'w')
    try:
        a.write(cnt % {'version': VERSION, 'fullversion':
                       FULLVERSION, 'isrelease': str(ISRELEASED)})
    finally:
        a.close()

local_path = os.path.dirname(os.path.abspath(sys.argv[0]))
os.chdir(local_path)
sys.path.insert(0, local_path)
sys.path.insert(0, os.path.join(local_path, 'kaleidoscope'))  # to retrive _version

# always rewrite _version
if os.path.exists('kaleidoscope/version.py'):
    os.remove('kaleidoscope/version.py')

write_version_py()


# Add command for running pylint from setup.py
class PylintCommand(setuptools.Command):
    """Run Pylint on all Kaleidoscope Python source files."""
    description = 'Run Pylint on Kaleidoscope Python source files'
    user_options = [
        # The format is (long option, short option, description).
        ('pylint-rcfile=', None, 'path to Pylint config file')]

    def initialize_options(self):
        """Set default values for options."""
        # Each user option must be listed here with their default value.
        self.pylint_rcfile = ''  # pylint: disable=attribute-defined-outside-init

    def finalize_options(self):
        """Post-process options."""
        if self.pylint_rcfile:
            assert os.path.exists(self.pylint_rcfile), (
                'Pylint config file %s does not exist.' % self.pylint_rcfile)

    def run(self):
        """Run command."""
        command = ['pylint']
        if self.pylint_rcfile:
            command.append('--rcfile=%s' % self.pylint_rcfile)
        command.append(os.getcwd()+"/kaleidoscope")
        subprocess.run(command, stderr=subprocess.STDOUT, check=False)


# Add command for running PEP8 tests from setup.py
class StyleCommand(setuptools.Command):
    """Run pep8 from setup."""
    description = 'Run style from setup'
    user_options = [
        # The format is (long option, short option, description).
        ('abc', None, 'abc')]

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        """Run command."""
        command = 'pycodestyle --max-line-length=100 kaleidoscope'
        subprocess.run(command, shell=True, check=False, stderr=subprocess.STDOUT)


setuptools.setup(
    name='kaleidoscope',
    version=VERSION,
    packages=PACKAGES,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    url="",
    author="Paul Nation",
    author_email="paul.nation@ibm.com",
    license="Apache 2.0",
    classifiers=[
        "Environment :: Web Environment",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering",
    ],
    cmdclass={'lint': PylintCommand, 'style': StyleCommand},
    install_requires=REQUIREMENTS,
    package_data=PACKAGE_DATA,
    include_package_data=True,
    zip_safe=False
)



# -*- coding: utf-8 -*-

# This code is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""General plotting utility functions"""


def hex_to_rgb(hex_color):
    """Converts a HEX color to a tuple of RGB values.

    Parameters:
        hex_color (str): Input hex color string.

    Returns:
        tuple: RGB color values.
    """
    hex_color = hex_color.lstrip("#")
    if len(hex_color) == 3:
        hex_color = hex_color * 2
    return int(hex_color[0:2], 16), int(hex_color[2:4], 16), int(hex_color[4:6], 16)


def find_text_color(hex_str):
    """Return correct text color, black or white,
    based on the background color.

    Parameters:
        hex_str (str): Hex color

    Returns:
        str: Output hex color for text
    """
    (r, g, b) = (hex_str[1:3], hex_str[3:5], hex_str[5:])
    color = "#ffffff"
    if 1 - (int(r, 16) * 0.299 + int(g, 16) * 0.587 + int(b, 16) * 0.114) / 255 < 0.5:
        color = '#000000'
    return color



# -*- coding: utf-8 -*-

# This file is part of Kaleidoscope.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=too-many-return-statements

# Modified from Qiskit version for better performance over
# wider range of values.

"""Local hardware information
"""
import platform
import psutil


def local_hardware_info():
    """Basic hardware information about the local machine.

    Gives actual number of CPU's in the machine, even when hyperthreading is
    turned on. CPU count defaults to 1 when true count can't be determined.

    Returns:
        dict: The hardware information.
    """
    results = {
        'os': platform.system(),
        'memory': psutil.virtual_memory().total / (1024 ** 3),
        'cpus': psutil.cpu_count(logical=False) or 1
    }
    return results



# -*- coding: utf-8 -*-

# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
# pylint: disable=broad-except, unspecified-encoding

"""Device layout information."""
import os
import json
import requests

# Try remote first since can be more up to date.
LAYOUTS = None
try:
    _REMOTE_JSON_URL = "https://github.com/nonhermitian/ \
                        ibm_quantum_system_layouts/raw/main/layouts.json"
    _req = requests.get(_REMOTE_JSON_URL)
    _req.raise_for_status()
except Exception:
    pass
else:
    LAYOUTS = _req.json()

if LAYOUTS is None:
    _LAYOUTS_DIR = os.path.dirname(os.path.realpath(__file__))
    with open(_LAYOUTS_DIR+"/layouts.json", 'r') as f:
        LAYOUTS = json.load(f)
    f.close()
